# -*- coding: latin-1 -*-


from .textmodel import texeltree
from .textmodel.textmodel import TextModel
from .textmodel.styles import create_style, updated_style, EMPTYSTYLE
from .textmodel.texeltree import Group, Text, groups, grouped, insert, \
    length, get_text, join, get_rightmost, NULL_TEXEL, dump, iter_leaves
from .textmodel.properties import overridable_property
from .wxtextview.boxes import Box, VGroup, VBox, Row, Rect, check_box, \
    NewlineBox, TextBox, TabulatorBox, extend_range_seperated, replace_boxes, \
    ChildBox, calc_length
from .wxtextview.simplelayout import create_paragraphs, Paragraph
from .wxtextview.testdevice import TESTDEVICE
from .wxtextview.builder import BuilderBase
from .wxtextview.wxdevice import WxDevice
from .wxtextview.wxtextview import WXTextView as _WXTextView

from .nbtexels import Cell, ScriptingCell, TextCell, Graphics, find_cell, \
    mk_textmodel, NotFound, strip_output, split_cell
from .clients import ClientPool
from .pyclient import PythonClient
from .nbstream import Stream, StreamRecorder
from .controlled import controlled
import string
import wx
import sys
import weakref




sepwidth = 20000 # a number which is just larger than the textwidth
wleft = None # width of left column, is set from NBView
scalefactor = None # scale line widths on HDPI-Displays, is set from
                   # NBView
tempstyle = dict(textcolor='blue', temp=True)


class ParagraphStack(VGroup):
    pass


class Frame(ChildBox):
    def __init__(self, childs, fillcolor=None, linecolor=None, border=(0, 0, 0, 0), 
                 device=None):
        self.border = border
        self.fillcolor = fillcolor or 'white'
        self.linecolor = linecolor or 'white'
        ChildBox.__init__(self, childs, device)

    def from_childs(self, childs):
        return Frame(childs, fillcolor, linecolor, border, self.device)

    def iter_boxes(self, i, x, y):
        border = self.border
        x += border[0]
        y += border[1]
        j1 = i
        for child in self.childs:
            j2 = j1+len(child)
            yield j1, j2, x, y, child
            y += child.height+child.depth
            j1 = j2

    def layout(self):
        w = h = d = 0
        border = self.border
        for child in self.childs:
            w += child.width
            h = max(h, child.height)
            d = max(d, child.depth)
        self.width = w+border[0]+border[2]
        self.height = h+border[1]+border[3]
        self.depth = d
        self.length = calc_length(self.childs)

    def get_index(self, x, y):
        l = [0, len(self)]
        for j1, j2, x1, y1, child in self.riter_boxes(0, 0, 0):
            if y1 <= y <= y1+child.height+child.depth:
                l.append(j1)
                i = child.get_index(x-x1, y-y1)
                if i is not None:
                    l.append(i+j1)
                if len(child):
                    l.append(j1+len(child))
                    l.append(j1+len(child)-1) # This is needed for rows
        return self._select_index(l, x, y)

    def draw(self, x, y, dc, styler):
        device = self.device
        dc.SetBrush(wx.Brush(self.fillcolor, wx.SOLID))
        dc.SetPen(wx.Pen(self.linecolor,style=wx.SOLID))
        dc.DrawRectangle(x, y, self.width, self.height)

        for j1, j2, x1, y1, child in self.iter_boxes(0, x, y):
            r = Rect(x1, y1, x1+child.width, y1+child.height)
            if device.intersects(dc, r):
                child.draw(x1, y1, dc, styler)



def draw_bracket(x, y, h, dc):
    dc.SetPen(wx.Pen("darkblue", 1*scalefactor))
    dc.DrawLines([
        (x+10, y),
        (x+6, y),
        (x+6, y+h-8),
        (x+10, y+h-8),
    ])
        


class TextCellBox(VBox):
    text = property(lambda s:s.childs[0])
    def __init__(self, textbox, device=None):
        # NOTE: textbox should be a PargraphStack
        if device is not None:
            self.device=device
        self.childs = [textbox]
        self.layout()

    def from_childs(self, childs):
        box = self.__class__(childs[0], self.device)
        return [box]

    def __len__(self):
        return self.length

    def create_group(self, l):
        return VGroup(l, device=self.device)

    def iter_boxes(self, i, x, y):
        text = self.text
        height = self.height
        j1 = i+1
        j2 = j1+len(text)
        yield j1, j2, x+wleft, y, text

    def layout(self):
        # compute w and h
        for j1, j2, x, y, child in self.iter_boxes(0, 0, 0):
            w = x+child.width
            h = y+child.height
            dh = y+child.height+child.depth
        self.width = w
        self.height = h
        self.depth = dh-h
        self.length = j2

    def __repr__(self):
        return self.__class__.__name__+'(%s)' % \
            (repr(self.text),)

    def extend_range(self, i1, i2):
        if i1<=0 or i2>=len(self):
            return 0, len(self)
        return extend_range_seperated(self, i1, i2) # XXX STIMMT DAS?

    def draw_selection(self, i1, i2, x, y, dc):
        if i1<=0 and i2>=self.length:
            self.device.invert_rect(x, y, sepwidth, self.height, dc)
        else:
            VBox.draw_selection(self, i1, i2, x, y, dc)

    def responding_child(self, i, x0, y0):
        # As default behaviour a VBox manages all index positions 0 to
        # n. Here we want position n to be managed by the next object. 
        if i == len(self):
            # Index i is between this box and the next box. We return
            # None to signal that the next box should be the
            # responding box.
            return None, i, x0, y0
        return VBox.responding_child(self, i, x0, y0)

    def get_cursorrect(self, i, x0, y0, style):
        child, j, x1, y1 = self.responding_child(i, x0, y0)
        if child is not None:
            return child.get_cursorrect(i-j, x1, y1, style)
        return self.get_rect(i, x0, y0)

    def get_rect(self, i, x0, y0):
        child, j, x1, y1 = self.responding_child(i, x0, y0)
        if child is not None:
            return child.get_rect(i-j, x1, y1)
        if i == 0:
            return Rect(x0, y0, sepwidth, y0+2)
        assert i == len(self)
        h = self.height
        return Rect(x0, y0+h, sepwidth, y0+h+2)

    def draw(self, x, y, dc, styler):
        draw_bracket(x, y, self.height, dc)
        VBox.draw(self, x, y, dc, styler)



class ScriptingCellBox(VBox):
    input = property(lambda s:s.childs[0])
    output = property(lambda s:s.childs[1])
    border = 8, 8, 8, 8
    inbackground = '#f7f7f7'
    inline = '#cfcfcf'
    promptstyle = create_style(
        textcolor = 'darkblue',
        #bold = True,
        italic = True,
        fontsize = 8,
        )


    def __init__(self, inbox, outbox, number=0, device=None):
        #assert isinstance(inbox, ParagraphStack)
        #assert isinstance(outbox, ParagraphStack)
        self.number = number
        if device is not None:
            self.device=device
        self.childs = [inbox, outbox]
        self.layout()

    def from_childs(self, childs):
        box = self.__class__(childs[0], childs[1], self.number, self.device)
        return [box]

    def __len__(self):
        return self.length

    def create_group(self, l):
        return VGroup(l, device=self.device)

    def iter_boxes(self, i, x, y):
        input = self.input
        output = self.output
        height = self.height        
        j1 = i+1
        j2 = j1+len(input)
        yield j1, j2, x+wleft, y, input
        j1 = j2
        j2 = j1+len(output)
        y += input.height+input.depth
        yield j1, j2, x+wleft, y, output

    def layout(self):
        # compute w and h
        dh = h = w = 0
        for j1, j2, x, y, child in self.iter_boxes(0, 0, 0):
            w = max(x+child.width, w)
            h = max(y+child.height, h)
            dh = max(y+child.height+child.depth, dh)
        self.width = w
        self.height = h
        self.depth = dh-h
        self.length = j2

    def __repr__(self):
        return self.__class__.__name__+'(%s, %s)' % \
            (repr(self.input),
             repr(self.output))

    def extend_range(self, i1, i2):
        for i in (0, len(self.input), len(self)-1):
            if i1<= i<i2:
                #print "empty in i1..i2", i, (i1, i2)
                #self.dump_boxes(0, 0, 0)
                return min(i1, 0), max(i2, len(self))
        return extend_range_seperated(self, i1, i2)

    def draw(self, x, y, dc, styler):
        draw_bracket(x, y, self.height, dc)
        a, b = list(self.iter_boxes(0, x, y))
        promptstyle = self.promptstyle
        styler.set_style(promptstyle)
        
        n = self.number or ''
        w1 = self.device.measure('Out', promptstyle)[0]
        w2 = self.device.measure('Out [%s]:' % n, promptstyle)[0]

        border = self.border
        x1 = x+wleft-w2
        x2 = x1+w1
        y1 = a[3]+border[1]
        y2 = b[3]+border[1]

        dc.DrawText("In", x1, y1)
        dc.DrawText("[%s]:" % n, x2, y1)
        if b[4].length > 1:
            dc.DrawText("Out", x1, y2)
            dc.DrawText("[%s]:" % n, x2, y2)

        VBox.draw(self, x, y, dc, styler)


    def draw_selection(self, i1, i2, x, y, dc):
        if i1<=0 and i2>=self.length:
            self.device.invert_rect(x, y, sepwidth, self.height, dc)
        else:
            VBox.draw_selection(self, i1, i2, x, y, dc)

    def responding_child(self, i, x0, y0):
        # Index position n+1 usually is managed by a child object. We
        # want the next object to be responsible, so we have to change
        # the return behaviour.
        if i == len(self):
            return None, i, x0, y0
        return VBox.responding_child(self, i, x0, y0)

    def get_cursorrect(self, i, x0, y0, style):
        child, j, x1, y1 = self.responding_child(i, x0, y0)
        if child is not None:
            return child.get_cursorrect(i-j, x1, y1, style)
        return self.get_rect(i, x0, y0)

    def get_rect(self, i, x0, y0):
        child, j, x1, y1 = self.responding_child(i, x0, y0)
        if child is not None:
            return child.get_rect(i-j, x1, y1)
        if i == 0:
            return Rect(x0, y0, sepwidth, y0+2)
        assert i == len(self)
        h = self.height
        return Rect(x0, y0+h, sepwidth, y0+h+2)





 


class BitmapBox(Box):
    def __init__(self, bitmap, device=None):
        if device is not None:
            self.device = device
        self.bitmap = bitmap
        self.width, self.height = bitmap.Size
        self.depth = 0

    def __len__(self):
        return 1

    def iter_boxes(self, i, x, y):
        if 0: yield 0,0,0

    def draw(self, x, y, dc, styler):
        dc.DrawBitmap(self.bitmap, x, y, useMask=False)

    def get_index(self, x, y):
        if x>self.width/2.0:
            return 1
        return 0

    def draw_selection(self, i1, i2, x, y, dc):
        self.device.invert_rect(x, y, self.width, self.height, dc)



def copy_pen(pen):
    new = wx.Pen(pen.Colour, pen.Width, pen.Style)
    new.Cap = pen.Cap
    new.Dashes = pen.Dashes
    new.Join = pen.Join
    return new


def copy_brush(brush):
    new = wx.Brush(brush.Colour, brush.Style)
    if brush.Stipple.Ok():
        new.Stipple = brush.Stipple
    return new


class GraphicsBox(Box):
    def __init__(self, texel, device=None):
        if device is not None:
            self.device = device
        self.texel = texel
        self.width, self.height = texel.size

    def __len__(self):
        return 1

    def iter_boxes(self, i, x, y):
        if 0: yield 0,0,0

    def draw(self, x, y, dc, styler):
        gc = wx.GraphicsContext.Create(dc)
        gc.Clip(x, y, self.width, self.height)
        gc.Translate(x, y)
        pen = wx.Pen(colour='black', style=wx.USER_DASH)
        brush = copy_brush(wx.TRANSPARENT_BRUSH)
        font = wx.SystemSettings.GetFont(wx.SYS_DEFAULT_GUI_FONT)
        matrix = gc.CreateMatrix() # transforms paths
        trafo = gc.GetTransform() # transforms the canvas
        state = dict(pen=pen, brush=brush, font=font, matrix=matrix,
                     trafo=trafo)
        gc.SetPen(pen)
        gc.SetBrush(brush)
        gc.SetFont(font)

        texel = self.texel
        if texel.frame:
            gc.DrawRectangle(1, 1, self.width-2, self.height-2)

        def draw(item, state=state, gc=gc):
            if type(item) is list or type(item) is tuple:
                oldbrush = copy_brush(state['brush'])
                oldpen = copy_pen(state['pen'])
                oldmatrix = gc.CreateMatrix(*state['matrix'].Get())
                gc.PushState()
                for child in item:
                    draw(child)
                gc.PopState()
                gc.SetPen(oldpen)
                gc.SetBrush(oldbrush)
                state['brush'] = oldbrush
                state['pen'] = oldpen
                state['matrix'] = oldmatrix
            else:
                item.draw(gc, state)

        try:
            draw(texel.items)
        except Exception, e:
            dc.SetBrush(wx.RED_BRUSH)
            dc.SetPen(wx.RED_PEN)
            dc.DrawRectangle(x, y, self.width, self.height)
            print >>sys.stderr, e
        draw = None
        
    def get_index(self, x, y):
        if x>self.width/2.0:
            return 1
        return 0

    def draw_selection(self, i1, i2, x, y, dc):
        self.device.invert_rect(x, y, self.width, self.height, dc)



class BoxesCache:
    def __init__(self):
        self.buffer = dict()
        
    def set(self, key, box):
        def callback(*args):
            del self.buffer[key]        
        self.buffer[key] = weakref.ref(box, callback)
        
    def get(self, key):
        return self.buffer[key]()

    def clear(self):
        self.buffer.clear()



def reset_numbers(texel):
    if isinstance(texel, ScriptingCell):
        texel.number = 0
    elif texel.is_group:
        for child in texel.childs:
            reset_numbers(child)


def is_temp(model, i):
    return model.get_style(i).get('temp', False)


def find_temp(tree):
    j1 = j2 = None
    for i1, i2, texel in iter_leaves(tree):
        if texel.style.get('temp'):
            if j1 is None:
                j1 = i1
                j2 = i2
            else:
                j2 = max(j2, i2)
    return j1, j2



class Builder(BuilderBase):
    parstyle = EMPTYSTYLE
    def __init__(self, model, clients=None, device=TESTDEVICE, maxw=0):
        if clients is None:
            clients = ClientPool()
        self._clients = clients
        self.device = device
        self.model = model
        self._maxw = maxw # width of page
        self.cache = BoxesCache()

    def get_device(self):
        return self.device

    def mk_style(self, style):
        # This can overriden e.g. to implement style sheets. The
        # default behaviour is to use the paragraph style and add the
        # text styles.
        r = self.parstyle.copy()
        r.update(style)
        return r

    def extended_texel(self):
        return self.model.get_xtexel()

    def maxw(self):
        return self._maxw

    def set_maxw(self, maxw):
        if maxw != self._maxw:
            self._maxw = maxw
            self.cache.clear()
            self.rebuild()

    ### Factory methods
    def create_boxes(self, texel):
        name = texel.__class__.__name__+'_handler'
        handler = getattr(self, name)
        #print "calling handler", name
        l = handler(texel)
        try:
            assert calc_length(l) == length(texel)
        except:
            print "handler=", handler
            raise
        return tuple(l)

    def Group_handler(self, texel):
        # Handles group texels. Note that the list of childs is
        # traversed from right to left. This way the "Newline" which
        # ends a line is handled before the content in the line. This
        # is important because in order to build boxes for the line
        # elements, we need to know the paragraph style which is
        # located in the NewLine-Texel.
        create_boxes = self.create_boxes
        r = ()
        for j1, j2, child in reversed(list(texeltree.iter_childs(texel))):
            r = create_boxes(child)+r
        return r

    def Text_handler(self, texel):
        # non caching version
        return [TextBox(texel.text, self.mk_style(texel.style), 
                self.device)]

    _textcache = dict()
    _textcache_keys = []
    def Text_handler(self, texel):
        # caching version
        key = texel.text, id(texel.style), id(self.parstyle), self.device
        try:
            return self._textcache[key]
        except: pass        
        r = [TextBox(texel.text, self.mk_style(texel.style), 
                     self.device)]
        self._textcache_keys.insert(0, key)        
        if len(self._textcache_keys) > 10000:
            _key = self._textcache_keys.pop()
            del self._textcache[_key]
        self._textcache[key] = r
        return r
    
    def NewLine_handler(self, texel):
        self.parstyle = texel.parstyle
        if texel.is_endmark:
            return [self.EndBox(self.mk_style(texel.style), self.device)]
        return [NewlineBox(self.mk_style(texel.style), self.device)] # XXX: Hmmmm

    def Tabulator_handler(self, texel):
        return [TabulatorBox(self.mk_style(texel.style), self.device)]

    def TextCell_handler(self, texel):
        try:
            cell = self.cache.get(texel)
        except KeyError:
            textbox = self.create_parstack(Group(texel.childs[1:]))
            cell = TextCellBox(textbox, device=self.device)
            self.cache.set(texel, cell)
        assert len(cell) == length(texel)
        return [cell]

    def ScriptingCell_handler(self, texel):
        sep1, inp, sep2, outp, sep3 = texel.childs

        border = ScriptingCellBox.border
        inbackground = ScriptingCellBox.inbackground
        inline = ScriptingCellBox.inline

        key = inp, 'input'
        try:
            inbox = self.cache.get(key)
        except KeyError:        
            client = self._clients.get_matching(texel)
            model = mk_textmodel(Group([inp, sep2]))

            t1, t2 = find_temp(inp)
            if t1 is not None:
                temp = model.remove(t1, t2)

            colorized = mk_textmodel(
                client.colorize(model.texel, bgcolor=inbackground))
            if t1 is not None:
                colorized.insert(t1, temp)
            inbox = self.create_parstack(colorized.texel)
            # change width so that boxes are always drawn with the
            # same width
            inbox.width = max(self._maxw-wleft, inbox.width)
            self.cache.set(key, inbox)
            assert len(inbox) == length(inp)+1

        key = outp, 'out'
        try:
            outbox = self.cache.get(key)
        except KeyError:
            outbox = self.create_parstack(Group([outp, sep3]))
            self.cache.set(key, outbox)
            assert len(outbox) == length(outp)+1

        cell = ScriptingCellBox(
            Frame([inbox], border=border, fillcolor=inbackground,
                  linecolor=inline, device=self.device), 
            Frame([outbox], border=border, device=self.device), 
            number=texel.number,
            device=self.device)
        return [cell]

    def Plot_handler(self, texel):
        return [PlotBox(device=self.device)]

    def Graphics_handler(self, texel):
        return [GraphicsBox(texel, device=self.device)]

    def BitmapRGB_handler(self, texel):
        w, h = texel.size
        bitmap = wx.BitmapFromBuffer(w, h, texel.data)
        return [BitmapBox(bitmap, device=self.device)]

    def BitmapRGBA_handler(self, texel):
        w, h = texel.size
        im = wx.ImageFromDataWithAlpha(w, h, texel.data, texel.alpha)
        bitmap = wx.BitmapFromImage(im)
        return [BitmapBox(bitmap, device=self.device)]

    ### Builder methods
    def create_paragraphs(self, texel):
        boxes = self.create_boxes(texel)
        if self._maxw:
            maxw = max(100, self._maxw-wleft)
        else:
            maxw = 0
        l = create_paragraphs(
            boxes, maxw = maxw,
            Paragraph = Paragraph,
            device = self.device)
        return l

    def create_parstack(self, texel):
        l = self.create_paragraphs(texel)
        return ParagraphStack(l, device=self.device)

    def rebuild(self):
        model = self.model
        boxes = self.create_boxes(model.texel)
        vgroup = VGroup(boxes, device=self.device)
        if vgroup.length:
            # Put a frame around the cells so that there is empty
            # space at the top and at the bottom.
            self._layout = Frame([vgroup], border=(0, 10, 0, 10),
                                 device=self.device)
        else:
            # Here we avoid VGroup(...) because I don't like how it
            # draws the caret. Might be a bug!??
            self._layout = vgroup

    ### Signal handlers
    def properties_changed(self, i1, i2):
        self.rebuild()

    def inserted(self, i, n):
        self.rebuild()

    def removed(self, i, n):
        #print "removed", i, n
        self.rebuild()




def common(s1, s2):
    # Helper: find the common part of two strings. Needed by
    # completer.
    r = ''
    i = 0
    try:
        while s1[i] == s2[i]:
            i += 1
    except IndexError:
        pass
    return s1[:i]


def strip_cells(texel):
    # Helper: replace all cells in texel by their content
    if texel.is_group:
        l = []
        for child in texel.childs:
            l.extend(strip_cells(child))
        return groups(l)
    if isinstance(texel, Cell):
        return join(*[[c] for c in texel.childs]) 
    return [texel]



class NBView(_WXTextView):
    temp_range = (0, 0)
    ScriptingCell = ScriptingCell
    _maxw = 0 # will be set later
    def __init__(self, parent, id=-1,
                 pos=wx.DefaultPosition, size=wx.DefaultSize, style=0, 
                 resize=False, filename=None, maxw=None):
        self.init_clients()
        self.do_resize = resize
        _WXTextView.__init__(self, parent, id=id, pos=pos, size=size,
                             style=style)
        
        # XXX Hackish method to set a width which scales with dpi
        global wleft
        global scalefactor
        if wleft is None:
            dc = wx.ScreenDC()
            dc.SetFont(self.Font)
            scalefactor = dc.CharWidth / float(dc.Font.GetPointSize())
            wleft = dc.GetTextExtent("123456789012345")[0]
            
        if maxw is None:
            dc = wx.ScreenDC()
            dc.SetFont(self.Font)
            maxw = dc.GetTextExtent("123456789012345")[0]*8
        self.maxw = maxw # note that this will also set builder.maxw

        if filename is not None:
            self._load(filename)
        self.actions[(wx.WXK_TAB, False, False)] = 'complete'
        self.actions[(wx.WXK_RETURN, True, False)] = 'execute'
        self.actions[(wx.WXK_RETURN, False, False)] = 'insert_newline_indented'
        self.actions[(2, True, False)] = 'split_cell'
        self.actions[(8, True, False)] = 'help'
  
    def _load(self, filename):
        s = open(filename, 'rb').read()
        import cerealizerformat        
        model = cerealizerformat.loads(s)
        reset_numbers(model.texel)
        self.set_model(model)

    @controlled
    def set_model(self, model):
        # Only meant to be called on fresh NBViews. Therefore we do
        # not reset cursor, selection etc.
        _WXTextView.set_model(self, model)

    def save(self, filename):        
        import cerealizerformat
        s = cerealizerformat.dumps(self.model)
        open(filename, 'wb').write(s)

    def init_clients(self):
        self._clients = ClientPool()        
        client = PythonClient()
        client.namespace['__shell__'] = self
        self._clients.register(client)
        # Hack to be able to load file created with older versions:
        self._clients.add(client, 'direct python')

        # Hack to make debugging easier
        self.pyclient = client

    _resize_pending = False
    _new_size = None
    def on_size(self, event):
        if not self.do_resize:
            return

        # Note that resize involves computing all line breaks and is
        # therefore a very costly operation. We therefore try to avoid
        # unnecessary resize events.
        self._new_size = event.Size
        if self._resize_pending:
            return
        self._resize_pending = True
        wx.CallAfter(self._adjust_size)
    
    def _adjust_size(self):
        self._resize_pending = False
        maxw = self._new_size[0]
        if maxw == self._maxw:
            return
        self.set_maxw(maxw)
        self.keep_cursor_on_screen()

    def create_builder(self):
        return Builder(
            self.model,
            clients=self._clients,
            device=WxDevice(),
            maxw=self._maxw)

    def print_temp(self, text):
        # Notes: 
        #  - temp has to be removed before other temp can be produced
        #  - temp does not generate undo information
        #  - temp must be removed before any undo operation
        if self.has_temp():
            raise Exception("Temp already set")
        properties = tempstyle.copy()
        properties['bgcolor'] = ScriptingCellBox.inbackground
        new = TextModel(text, **properties)
        i = self.index
        self.model.insert(i, new)
        j1, j2 = self.temp_range
        if j1 == j2:
            self.temp_range = i, i+len(new)
        else:
            self.temp_range = j1, i+len(new)

    def clear_temp(self):
        self.complete_count = 0
        j1, j2 = self.temp_range
        if j1 != j2:
            self.model.remove(j1, j2)
            self.temp_range = j1, j1

    def has_temp(self):
        j1, j2 = self.temp_range
        return not j1 == j2

    def get_word(self, j):
        # get word which ends at *j*
        model = self.model
        row, col = model.index2position(j)
        text = model.get_text(model.linestart(row), j)
        i = len(text)-1
        while i>=0 and text[i] in string.letters+string.digits+"_.":
            i -= 1
        if j == i:
            return ''
        return text[i+1:]

    complete_count = 0
    def handle_action(self, action, shift=False, memo=None):
        complete_count = self.complete_count
        self.clear_temp()
        if action == 'complete_or_help':
            if complete_count > 0:
                action = 'help'
            else:
                action = 'complete'

        if action == 'split_cell':
            return self.split_cell()
        elif action == 'execute':
            try:
                self.execute()
            except NotFound:
                pass
            return
        if action not in ('complete', 'help'):
            return _WXTextView.handle_action(self, action, shift)
        # complete or help#
        try:
            i0, cell = self.find_cell()
        except NotFound:
            return
        
        if not isinstance(cell, ScriptingCell):
            return        
        index = self.index
        if index <= i0 or index > i0+length(cell.input)+1:
            return
        
        word = self.get_word(index)
        client = self._clients.get_matching(cell)
        
        if client is None:
            return
        
        if action == 'help':
            s = client.help(word)
            self.print_temp('\n'+s+'\n')

        elif action == 'complete':
            maxoptions = 50
            options = client.complete(word, maxoptions)
            if not options or list(options) == [word]:
                self.print_temp( "\n[No completion]\n")
                self.complete_count = 1
            else:
                completion = reduce(common, options)[len(word):]
                if completion and len(options) != maxoptions:
                    self.insert_text(index, completion)
                    index += len(completion)
                else:
                    options = list(sorted(options))
                    s = ', '.join(options)
                    if len(options) == maxoptions:
                        s += ' ... '
                    self.print_temp('\n'+s+'\n')
        self.index = index
        self.adjust_viewport()
        
    def on_char(self, event):
        keycode = event.GetKeyCode()
        ctrl = event.ControlDown()
        shift = event.ShiftDown()
        if keycode == wx.WXK_RETURN and shift and not ctrl:
            try:
                self.execute()
            except NotFound:
                pass
        else:
            _WXTextView.on_char(self, event)

    def find_cell(self):
        return find_cell(self.model.texel, self.index)

    @controlled
    def execute(self):
        self.clear_temp()
        try:
            i0, cell = self.find_cell()
        except NotFound:
            return
        if not isinstance(cell, ScriptingCell):
            self.index = i0+length(cell)
            return
        n = length(cell)
        client = self._clients.get_matching(cell)
        stream = Stream()
        result = client.execute(cell.childs[1], stream.output)
        new = self.ScriptingCell(cell.childs[1], stream.model.texel, 
                                 client.counter)
        assert i0>=0
        assert i0+n<=len(self.model)
        infos = []
        infos.append(self._remove(i0, i0+n))
        self.model.insert(i0, mk_textmodel(new))
        infos.append((self._remove, i0, i0+length(new)))
        self.add_undo(infos) 
        self.adjust_viewport()
        return result

    def execute_all(self):
        self.index = 0
        while 1:
            try:
                i, cell = self.find_cell()
            except NotFound:
                break
            self.execute()

    @controlled
    def reset_interpreter(self):
        self.init_clients()

    def insert(self, i, textmodel):
        insidecell = False
        try:
            i0, cell = self.find_cell()
            if i0 < i < i0+length(cell):
               insidecell = True
        except NotFound:
            pass
        try:
            find_cell(textmodel.texel, 0)
            hascells = True
        except NotFound:
            hascells = False
        if insidecell:
            if hascells:
                # We can't insert cells inside cells. Therefore we
                # remove all outer cells from what we insert.
                texel = grouped(strip_cells(textmodel.texel))
                _WXTextView.insert(self, i, mk_textmodel(texel))
            else:
                _WXTextView.insert(self, i, textmodel)
        else:
            if hascells:
                _WXTextView.insert(self, i, textmodel)
            else:
                cell = self.ScriptingCell(NULL_TEXEL, NULL_TEXEL)
                _WXTextView.insert(self, i, mk_textmodel(cell))
                _WXTextView.insert(self, i+1, textmodel)

    def undo(self):
        self.clear_temp()
        _WXTextView.undo(self)
                
    def redo(self):
        self.clear_temp()
        _WXTextView.redo(self)

    @controlled
    def remove_output(self):
        self.transform(strip_output)

    @controlled
    def split_cell(self):
        i = self.index
        self.transform(lambda texel, i=i:split_cell(texel, i))

    def between_cells(self):
        i = self.index
        insidecell = False
        try:
            i0, cell = self.find_cell()
            if i0 < i < i0+length(cell):
               insidecell = True
        except NotFound:
            pass
        return not insidecell

    can_insert_textcell = can_insert_pycell = between_cells

    @controlled
    def insert_textcell(self):
        "Insert text cell"
        cell = TextCell(NULL_TEXEL)
        i = self.index
        self.insert(i, mk_textmodel(cell))
        self.index = i+1

    @controlled
    def insert_pycell(self):
        "Insert python cell"
        cell = ScriptingCell(NULL_TEXEL, NULL_TEXEL)
        i = self.index
        self.insert(i, mk_textmodel(cell))
        self.index = i+1

    set_index = controlled(_WXTextView.set_index)
    set_selection = controlled(_WXTextView.set_selection)




def init_testing(redirect=True):
    app = wx.App(redirect=redirect)
    model = TextModel('')

    frame = wx.Frame(None)
    win = wx.Panel(frame)
    view = NBView(win)
    view.model = model
    box = wx.BoxSizer(wx.VERTICAL)
    box.Add(view, 1, wx.ALL|wx.GROW, 1)
    win.SetSizer(box)
    win.SetAutoLayout(True)
    
    frame.Show()
    return locals()

def test_00():
    "cell"
    ns = init_testing(False)
    cell = ScriptingCell(Text(u'1234567890'), Text(u'abcdefghij'))
    assert length(cell.input) == 10
    assert length(cell.output) == 10
    assert length(cell) == 23

    texel = grouped(insert(cell, 1, [Text(u'x')]))
    assert get_text(texel)[1:2] == u'x'


def test_02():
    "execute"
    return # XXX skip this for now
    ns = init_testing(False)
    cell = ScriptingCell(Text(u'1+2'), Text(u''))
    cell = cell.execute()
    assert get_text(cell.output) == '3'

    cell = ScriptingCell(Text(u'for a in range(2):\n    print a'),
                         Text(u''))
    cell = cell.execute()
    assert get_text(cell.output) == u'0\n1\n'

    cell = ScriptingCell(Text(u'asdsad'), Text(u''))
    cell = cell.execute()
    
    assert get_text(cell.output) == u'  File "In[3]", line 1, ' \
        'in <module>\nNameError: name \'asdsad\' is not defined\n'

def test_03():
    "find_cell"
    tmp1 = TextModel(u'for a in range(3):\n    print a')
    tmp2 = TextModel(u'for a in range(10):\n    print a')
    cell1 = ScriptingCell(tmp1.texel, Text(u''))
    cell2 = ScriptingCell(tmp2.texel, Text(u''))

    model = TextModel('')
    model.insert(len(model), mk_textmodel(cell1))
    model.insert(len(model), mk_textmodel(cell2))

    assert find_cell(model.texel, 1) == (0, cell1)
    assert find_cell(model.texel, length(cell1)-1) == (0, cell1)

    assert find_cell(model.texel, length(cell1)) == (length(cell1), cell2)
    assert find_cell(model.texel, length(cell1)+5) == (length(cell1), cell2)


def test_04():
    "copy cells"
    model = TextModel('')
    tmp = TextModel(u'for a in range(5):\n    print a')
    cell = ScriptingCell(tmp.texel, Text(u''))
    model.insert(len(model), mk_textmodel(cell))
    tmp = model.copy(0, len(model))
    model.insert(0, tmp)


def test_05():
    "CellBox"
    empty = ParagraphStack([])
    cell1 = ScriptingCellBox(empty, empty)
    check_box(cell1.input)
    check_box(cell1)
    #cell1.dump_boxes(0, 0, 0)

    cell2 = ScriptingCellBox(empty, empty)
    check_box(cell2)
    stack = VGroup([cell1, cell2])
    check_box(stack)

    stack = VGroup([])
    check_box(stack)

    t1 = TextBox("0123456789")
    t2 = TextBox("abcdefghij")
    t3 = TextBox("xyz")
    NL = NewlineBox()
    p1 = Paragraph([Row([t1, NL])])
    cell1 = ScriptingCellBox(
        ParagraphStack([Paragraph([Row([t1, NL])]), Paragraph([Row([t2, NL])]),]), 
        ParagraphStack([Paragraph([Row([t3, NL])])]))

    (j1, j2, inp), (k1, k2, outp) = cell1.iter_childs()
    assert (j1, j2) == (1, 23) # input box

    cell2 = ScriptingCellBox(ParagraphStack([Row([t3, NL])]), empty)

    g = VGroup([cell1, cell2])


def test_10():
    "Factory"
    ns = init_testing(False)
    cell = ScriptingCell(Text(u'a'), Text(u'b'))
    factory = Builder(TextModel(''))
    factory._clients.register(PythonClient())
    boxes = factory.create_boxes(cell)
    assert len(boxes) == 1
    cellbox = boxes[0]
    assert len(cellbox) == 5
    assert length(cell) == 5

    check_box(cellbox)
    check_box(cellbox.input)
    check_box(cellbox.output)
    return ns


def test_11():
    ns = init_testing(redirect=False)
    model = ns['model']
    model.remove(0, len(model))
    tmp = TextModel(u'for a in range(16):\n    print a')
    cell = ScriptingCell(tmp.texel, Text(u''))
    model.insert(len(model), mk_textmodel(cell))

    assert model.index2position(0) == (0, 0)
    assert model.index2position(1) == (1, 0)
    cell = model.texel

    assert find_cell(model.texel, 1) == (0, cell)

    view = ns['view']
    view.index = 1
    #print model.texel
    view.execute()

    check_box(view.builder._layout, model.texel)

    #view.layout.dump_boxes(0, 0, 0)
    #assert inside_cell(view.layout, 68, 68)

    #texeltree.dump(model.texel)

    model.insert_text(68, u'x')
    assert model.get_text()[65:71] == '14\nx15'
    #view.layout.dump_boxes(0, 0, 0)
    
    model.remove(68, 69)
    #view.layout.dump_boxes(0, 0, 0)
    assert model.get_text()[65:70] == '14\n15'

    model.insert(0, mk_textmodel(cell))
    model.remove(0, length(cell))

    # insert new cell, insert input
    view.insert(0, TextModel("a=1"))

    # insert output
    view.insert(5, TextModel("x"))

    # remove the complete input-part
    model.remove(1, 4)

    # remove the complete output-part
    model.remove(2,3)
    
    return ns

def test_12():
    ns = init_testing(True)
    model = ns['model']
    model.remove(0, len(model))
    tmp = TextModel(u'for a in range(5):\n    print a')
    cell = ScriptingCell(tmp.texel, NULL_TEXEL)
    model.insert(len(model), mk_textmodel(cell))

    assert find_cell(model.texel, 1) == (0, cell)

    view = ns['view']
    view.index = 1
    return ns


def test_13():
    ns = init_testing(True)
    model = ns['model']
    model.remove(0, len(model))
    tmp = TextModel(u'''import matplotlib.pyplot as plt
figure = plt.figure(
    facecolor='white',
    figsize=(3, 2.5))
figure.set_frameon(False)
plot = figure.add_subplot ( 111 )
plt.plot([1,2,3,4])
plt.ylabel('some numbers')
output(figure)''')
    cell = ScriptingCell(tmp.texel, NULL_TEXEL)
    model.insert(len(model), mk_textmodel(cell))

    assert find_cell(model.texel, 1) == (0, cell)

    view = ns['view']
    view.index = 1
    view.execute()
    return ns


def test_14():
    "cell cursor"
    ns = init_testing(False)
    model = ns['model']
    model.remove(0, len(model))
    tmp = TextModel(u'for a in range(5):\n    print a')
    cell = ScriptingCell(tmp.texel, NULL_TEXEL)
    model.insert(len(model), mk_textmodel(cell))

    assert find_cell(model.texel, 1) == (0, cell)

    view = ns['view']
    view.index = 1
    layout = view.builder.get_layout()
    r1 = layout.get_rect(0, 0, 0)
    assert r1.x2-r1.x1 == sepwidth

    r2 = layout.get_cursorrect(0, 0, 0, {})
    r3 = layout.get_cursorrect(len(model), 0, 0, {})
    assert r3.x2-r3.x1 == sepwidth
    assert r3.y1 > r2.y1


def test_15():
    "get_word, print_temp, remove_temp"
    ns = init_testing(False)
    model = ns['model']
    tmp = TextModel(u'for a in range(5):\n    print a')
    cell = ScriptingCell(tmp.texel, NULL_TEXEL)
    model.insert(len(model), mk_textmodel(cell))
    view = ns['view']
    assert view.get_word(12) == 'ra'

    text = model.get_text()
    view.index = 12
    view.print_temp('\n[raise range raw_input]\n')
    #print model.get_text()

    view.clear_temp()
    assert model.get_text() == text


def test_16():
    "Graphics"
    ns = init_testing(False)
    model = ns['model']
    model.insert(len(model), mk_textmodel(Graphics([])))
    view = ns['view']


def test_17():    
    "parstyle"
    model = TextModel('')
    tmp = TextModel(u'Line 1\nLine 2\nLine 3')
    cell = TextCell(tmp.texel)
    model.insert(len(model), mk_textmodel(cell))
    i0 = 0
    i1 = model.linestart(1)
    i2 = model.linestart(2)
    #print i0, i1, i2
    assert model.get_parstyle(i0) == {}
    assert model.get_parstyle(i1) == {}
    assert model.get_parstyle(i2) == {}
    #dump(model.texel)
    #print
    model.set_parproperties(i1, i2, bullet = True)
    #dump(model.texel)
    assert model.get_parstyle(i0) == {}
    assert model.get_parstyle(i1) == {'bullet':True}
    assert model.get_parstyle(i2) == {}
    

def demo_00():
    from .wxtextview import testing
    ns = test_11()
    testing.pyshell(ns)
    ns['app'].MainLoop()


def demo_01():
    from .wxtextview import testing
    ns = test_13()
    testing.pyshell(ns)
    ns['app'].MainLoop()





# -*- coding: latin-1 -*-

import wx


defaultstyle = dict(
    fontsize=10, bgcolor='white', textcolor='black', 
    underline=False, facename='', italic=False, bold=False)


def filled(style, defaultstyle=defaultstyle):
    # fill empty properties with default values
    new = defaultstyle.copy()
    new.update(style)
    return new


def invert_rect_INV(self, x, y, w, h, dc):
    if 1:
        dc.SetLogicalFunction(wx.INVERT)
        dc.SetBrush(wx.BLACK_BRUSH)
        dc.SetPen(wx.TRANSPARENT_PEN)
    else:
        # Alternative which gives a light blue selection: 
        dc.SetLogicalFunction(wx.XOR)
        dc.SetBrush(wx.RED_BRUSH)
    dc.DrawRectangle(x, y, w, h)


def invert_rect_BLIT(self, x, y, w, h, dc):
    dc.Blit(x, y, w, h, dc, x, y, wx.SRC_INVERT)

    
def invert_rect_FALLBACK(self, x, y, w, h, dc):
    gcdc = wx.GCDC(dc)
    r = g = 0
    b = 128    
    color = wx.Colour(r, g, b, 50)
    gcdc.SetBrush(wx.Brush(color))
    gcdc.SetPen(wx.TRANSPARENT_PEN)
    gcdc.DrawRectangle(x, y, w, h)

    
def get_font(style):
    weight = {False : wx.FONTWEIGHT_NORMAL,
              True : wx.FONTWEIGHT_BOLD}[style.get('bold', False)]
    slant = {False : wx.FONTSTYLE_NORMAL,
              True : wx.FONTSTYLE_ITALIC}[style.get('italic', False)]

    family = dict(
        roman = wx.FONTFAMILY_ROMAN,
        modern = wx.FONTFAMILY_MODERN, 
        swiss = wx.FONTFAMILY_SWISS,        
    )[style.get('family', 'modern')]
    return wx.Font(
        style['fontsize'], family, slant, weight,
        style['underline'], style['facename'])


# Profiling revelead that a lot of time is spent in the measure_*
# functions. We therefore cache the results in a fixed size dict. This
# reduces the time spent by ~90%.
_cache = dict()
_cache_keys = []

def measure_win(self, text, style):
    key = text, tuple(style.items())
    try:
        return _cache[key]
    except:
        pass
    style = filled(style)
    font = get_font(style)
    dc = wx.MemoryDC()
    bmp = wx.EmptyBitmap(1,1)
    dc.SelectObject(bmp)
    dc.SetFont(font)
    w, h = dc.GetTextExtent(text)
    _cache[key] = w, h
    _cache_keys.insert(0, key)
    if len(_cache_keys) > 1000:
        _key = _cache_keys.pop()
        del _cache[key]        
    return w, h

def measure_mac(self, text, style):
    key = text, tuple(style.items())
    try:
        return _cache[key]
    except:
        pass
    style = filled(style)
    font = get_font(style)
    gc = wx.GraphicsContext_CreateMeasuringContext()
    gc.SetFont(font)
    w, h = gc.GetTextExtent(text)
    _cache[key] = w, h
    _cache_keys.insert(0, key)
    if len(_cache_keys) > 1000:
        _key = _cache_keys.pop()
        del _cache[key]        
    return w, h

def measure_gtk(self, text, style):
    key = text, tuple(style.items())
    try:
        return _cache[key]
    except:
        pass
    style = filled(style)
    text = text.replace('\n', ' ') # replace newlines to avoid double lines
    font = get_font(style)
    # GC returns wrong font metric values in gtk! We therefore use the DC.  
    dc = wx.MemoryDC()
    bmp = wx.EmptyBitmap(1,1)
    dc.SelectObject(bmp)
    dc.SetFont(font)
    w, h = dc.GetTextExtent(text)
    _cache[key] = w, h
    _cache_keys.insert(0, key)
    if len(_cache_keys) > 1000:
        _key = _cache_keys.pop()
        del _cache[key]        
    return w, h


def measure_parts_win(self, text, style):
    style = filled(style)
    font = get_font(style)
    dc = wx.MemoryDC()
    bmp = wx.EmptyBitmap(1,1)
    dc.SelectObject(bmp)
    dc.SetFont(font)
    return dc.GetPartialTextExtents(text)


def measure_parts_gtk(self, text, style):
    style = filled(style)
    font = get_font(style)
    dc = wx.MemoryDC()
    bmp = wx.EmptyBitmap(1,1)
    dc.SelectObject(bmp)
    dc.SetFont(font)
    return dc.GetPartialTextExtents(text)


def measure_parts_mac(self, text, style):
    style = filled(style)
    font = get_font(style)
    gc = wx.GraphicsContext_CreateMeasuringContext()
    gc.SetFont(font)
    return gc.GetPartialTextExtents(text)


class WxDevice:
    def intersects(self, dc, rect):
        r = wx.Rect(rect.x1, rect.y1, rect.x2-rect.x1, rect.y2-rect.y1)
        return dc.ClippingRect.Intersects(r)

    def draw_text(self, text, x, y, dc):
        dc.DrawText(text, x, y)

    if "msw" in wx.version():
        invert_rect = invert_rect_BLIT
        measure = measure_win
        measure_parts = measure_parts_win
        buffering = True
    elif "gtk" in wx.version():
        invert_rect = invert_rect_BLIT
        #invert_rect = invert_rect_INV
        measure = measure_gtk
        measure_parts = measure_parts_gtk
        buffering = True
    elif 'mac' in wx.version():
        invert_rect = invert_rect_INV
        measure = measure_mac
        measure_parts = measure_parts_mac
        buffering = False
    invert_rect = invert_rect_FALLBACK

    

class DCStyler:
    last_style = None
    def __init__(self, dc):
        self.dc = dc
        
    def set_style(self, style):
        if style is self.last_style:
            return
        self.last_style = style

        _style = filled(style)
        font = get_font(_style)
        self.dc.SetFont(font)            
        try: # Phoenix
            self.dc.SetTextBackground(wx.Colour(_style['bgcolor']))
            self.dc.SetTextForeground(wx.Colour(_style['textcolor']))
        except TypeError: # Classic
            self.dc.SetTextBackground(wx.NamedColour(_style['bgcolor']))
            self.dc.SetTextForeground(wx.NamedColour(_style['textcolor']))

 



# -*- coding: latin-1 -*-


from ..textmodel import texeltree
from ..textmodel.textmodel import TextModel
from ..textmodel.texeltree import NewLine, Group, Text, length
from ..textmodel.styles import EMPTYSTYLE
from .testdevice import TESTDEVICE
from .boxes import TextBox, NewlineBox, TabulatorBox, EmptyTextBox, \
    EndBox, check_box, Box, calc_length



class Factory:
    # The Factory object takes a texeltree as an outline to create a
    # sequence of boxes. For every Texel-Class Factory has a
    # corresponding handler method of the same name,
    # e.g. Group_handler for Group-texels. The call signature is
    # handler(texel, i1, i2), where i1 and i2 denotes the part of
    # texel for which boxes are generated.

    TextBox = TextBox
    NewlineBox = NewlineBox
    TabulatorBox = TabulatorBox
    EndBox = EndBox

    parstyle = EMPTYSTYLE

    def __init__(self, device=TESTDEVICE):
        self.device = device

    def get_device(self):
        return self.device

    def mk_style(self, style):
        # This can overriden e.g. to implement style sheets. The
        # default behaviour is to use the paragraph style and add the
        # text styles.
        r = self.parstyle.copy()
        r.update(style)
        return r

    ### Factory methods
    def create_all(self, texel):
        # Convenience method
        return self.create_boxes(texel, 0, length(texel))

    def create_boxes(self, texel, i1, i2):
        assert i1>=0
        assert i2<=length(texel)
        assert i1<=i2
        if i1 == i2:
            return () # XXX Why is this needed?
        name = texel.__class__.__name__+'_handler'
        handler = getattr(self, name)
        #print "calling handler", name, i1, i2
        l = handler(texel, i1, i2)
        try:
            assert calc_length(l) == i2-i1
        except:
            print "handler=", handler
            raise
        return tuple(l)
        
    def Group_handler(self, texel, i1, i2):
        # Handles group texels. Note that the list of childs is
        # traversed from right to left. This way the "Newline" which
        # ends a line is handled before the content in the line. This
        # is important because in order to build boxes for the line
        # elements, we need the paragraph style which is located in
        # the NewLine-Texel.
        r = ()
        for j1, j2, child in reversed(list(texeltree.iter_childs(texel))):
            if i1 < j2 and j1 < i2: # overlapp
                r = self.create_boxes(child, max(0, i1-j1), min(i2, j2)-j1)+r
        return r

    def Text_handler(self, texel, i1, i2):
        return [self.TextBox(texel.text[i1:i2], self.mk_style(texel.style), 
                             self.device)]

    _cache = dict()
    _cache_keys = []
    def Text_handler(self, texel, i1, i2):
        # cached version
        key = texel.text, id(texel.style), id(self.parstyle), i1, i2, self.device
        try:
            return self._cache[key]
        except: pass        
        r = [self.TextBox(texel.text[i1:i2], self.mk_style(texel.style), 
                          self.device)]
        self._cache_keys.insert(0, key)        
        if len(self._cache_keys) > 10000:
            _key = self._cache_keys.pop()
            del self._cache[_key]
        self._cache[key] = r
        return r

    def NewLine_handler(self, texel, i1, i2):
        self.parstyle = texel.parstyle
        if texel.is_endmark:
            return [self.EndBox(self.mk_style(texel.style), self.device)]
        return [self.NewlineBox(self.mk_style(texel.style), self.device)] # XXX: Hmmmm

    def Tabulator_handler(self, texel, i1, i2):
        return [self.TabulatorBox(self.mk_style(texel.style), self.device)]






class BuilderBase:
    """The builder is responsible for creating and updating the layout. 

    The methods build, insert, remove and update return a tree of boxes (=
    "layout").

    The length of the box tree is always the length of the model +1,
    because we add a special box ("end mark").

    """

    _layout = None

    def rebuild(self):
        # sets self._layout
        pass

    def get_layout(self):
        assert self._layout is not None
        return self._layout

    ### Signal handlers
    def properties_changed(self, i1, i2):
        pass

    def inserted(self, i, n):
        pass

    def removed(self, i, n):
        pass



def test_01():
    factory = Factory()
    boxes = factory.create_all(TextModel("123").texel)
    assert calc_length(boxes) == 3
    boxes = factory.create_all(TextModel("123\n567").get_xtexel())
    assert calc_length(boxes) == 8

    





# -*- coding: latin-1 -*-


import wx
import  wx.lib.colourselect as  csel
import  wx.lib.rcsizer  as rcs

from .textmodel.viewbase import ViewBase
from .textmodel.texeltree import EMPTYSTYLE
from .textmodel.styles import create_style
from .wxtextview.wxdevice import defaultstyle



class Inspector(wx.Frame, ViewBase):
    def __init__(self, *args, **kwds):
        ViewBase.__init__(self)
        wx.Frame.__init__(self, *args, title='Text format',
                          style=wx.DEFAULT_FRAME_STYLE|wx.FRAME_FLOAT_ON_PARENT
                               |wx.FRAME_TOOL_WINDOW, **kwds)
        sizer1 = wx.BoxSizer( wx.VERTICAL )
        panel = wx.Panel(self)
        panel.SetBackgroundColour(wx.SystemSettings.GetColour(wx.SYS_COLOUR_BTNFACE))
        self.SetBackgroundColour(panel.BackgroundColour)
        sizer2 = wx.BoxSizer(wx.VERTICAL)

        e = wx.FontEnumerator()
        e.EnumerateFacenames()
        faces = sorted(e.GetFacenames())
        self.faces = ['']+faces
        self.fontface = wx.Choice(
            panel, choices=['<default font>']+faces)
        self.fontface.Bind(wx.EVT_CHOICE, self.on_face)

        sizer2.Add(self.fontface, 0, wx.ALL|wx.EXPAND, 5)
        sizer3 = wx.BoxSizer(wx.VERTICAL)

        choices = map(str, (8, 9, 10, 12, 14, 16, 18, 20, 22, 24, 26, 30))

        fun = lambda e:self.set_properties(fontsize=int(self.size.GetValue()))
        self.size = wx.ComboBox(
            panel, value = "12", choices=choices,
            style = wx.CB_DROPDOWN| wx.TE_PROCESS_ENTER
        )
        self.size.SetMinSize((100, -1))
        for binder in wx.EVT_TEXT_ENTER, wx.EVT_KILL_FOCUS, wx.EVT_COMBOBOX:
            self.size.Bind(binder, fun)
        sizer3.Add(self.size, 0, wx.ALL|wx.EXPAND, 5)

        sizer4 = wx.BoxSizer(wx.HORIZONTAL )
        self.fgcolor = csel.ColourSelect(panel, -1)
        sizer4.Add(self.fgcolor, 1, wx.ALL, 5)
        self.fgcolor.Bind(
            csel.EVT_COLOURSELECT,
            lambda e:self.set_properties(
                textcolor=self.fgcolor.GetValue().GetAsString()))

        self.bgcolor = csel.ColourSelect(panel, -1)
        sizer4.Add(self.bgcolor, 1, wx.ALL, 5)
        self.bgcolor.Bind(
            csel.EVT_COLOURSELECT,
            lambda e:self.set_properties(
                bgcolor=self.bgcolor.GetValue().GetAsString()))

        sizer3.Add(sizer4, 0, wx.EXPAND, 5)

        self.underline = wx.CheckBox(panel, -1, "Underline")
        sizer3.Add( self.underline, 0, wx.ALL|wx.EXPAND, 5 )
        self.underline.Bind(
            wx.EVT_CHECKBOX,
            lambda e:self.set_properties(underline=self.underline.GetValue()))

        self.bold = wx.CheckBox(panel, -1, "Bold")
        sizer3.Add( self.bold, 0, wx.ALL|wx.EXPAND, 5 )
        self.bold.Bind(
            wx.EVT_CHECKBOX,
            lambda e:self.set_properties(bold=self.bold.GetValue()))

        self.italic = wx.CheckBox(panel, -1, "Italic")
        sizer3.Add( self.italic, 0, wx.ALL|wx.EXPAND, 5 )
        self.italic.Bind(
            wx.EVT_CHECKBOX,
            lambda e:self.set_properties(italic=self.italic.GetValue()))

        button = wx.Button(panel, -1, "Clear")
        button.Bind(wx.EVT_BUTTON, self.on_clear)
        sizer3.Add( button, 0, wx.ALL, 5 )


        sizer2.Add( sizer3, 1, wx.ALL|wx.EXPAND, 5 )

        panel.SetSizer( sizer2 )
        panel.Layout()
        sizer2.Fit( panel )
        sizer1.Add( panel, 1, wx.EXPAND |wx.ALL, 5 )

        self.SetSizer( sizer1 )
        self.Layout()
        sizer1.Fit( self )


    def index_changed(self, textview):
        self.update()

    def selection_changed(self, textview):
        self.update()

    def on_clear(self, event):
        textview = self.model
        i1, i2 = sorted(textview.selection)
        textview.clear_styles(i1, i2)

    def on_face(self, event=None):
        i = self.fontface.Selection
        facename = self.faces[i]
        self.set_properties(facename = facename)

    def set_properties(self, **properties):
        textview = self.model
        i1, i2  = sorted(textview.selection)
        textview.set_properties(i1, i2, **properties)

    def update(self):
        textview = self.model
        index = textview.index
        textmodel = textview.model
        style = defaultstyle.copy()
        style.update(textmodel.get_parstyle(index))
        style.update(textmodel.get_style(max(0, index-1)))
        face = style['facename']
        i = self.faces.index(face)
        self.fontface.Selection = i
        self.fgcolor.SetValue(style['textcolor'])
        self.bgcolor.SetValue(style['bgcolor'])
        self.underline.SetValue(style['underline'])
        self.italic.SetValue(style['italic'])
        self.bold.SetValue(style['bold'])
        self.size.SetValue(str(style['fontsize']))


class ParagraphSettings:
    title = 'Paragraph'
    style = EMPTYSTYLE
    bullets = False
    enumeration = False # is this an enumerated element?
    section = False     # is this a section change?
    align = 'left'


class HeadingSettings(ParagraphSettings):
    title = 'Heading'
    section = True
    style = create_style(
        fontsize = 16,
        facename = 'Courier',
        weight = 'bold'
        )


class ListSettings(ParagraphSettings):
    title = 'List'
    enumeration = True


def demo_00():
    from .nbview import NBView
    from .nbtexels import TextCell, NULL_TEXEL, mk_textmodel
    app = wx.App(redirect=True)
    frame = wx.Frame(None)
    win = wx.Panel(frame)
    view = NBView(win)
    cell = TextCell(NULL_TEXEL)
    view.model.insert(0, mk_textmodel(cell))
    box = wx.BoxSizer(wx.VERTICAL)
    box.Add(view, 1, wx.ALL|wx.GROW, 1)
    win.SetSizer(box)
    win.SetAutoLayout(True)
    inspector = Inspector(None)
    inspector.model = view
    inspector.Show()
    
    frame.Show()
    app.MainLoop()



# -*- coding: latin-1 -*-


from copy import copy as shallow_copy


debug = 0
nmax = 15

def set_nmax(n):
    """Sets *nmax* to the value *i*.

       The value nmax determines the number of childs per group (which
       should be between nmax/2 and nmax for efficient trees).
    """
    global nmax
    nmax = n

EMPTYSTYLE = {}
style_pool = {():EMPTYSTYLE}

def hash_style(style):
    return tuple(sorted(style.items()))


def as_style(d):
    global style_pool
    key = hash_style(d)
    try:
        return style_pool[key]
    except KeyError:
        style_pool[key] = d
        return d


# ---- Tree objects ----
class Texel:
    is_single = 0
    is_container = 0
    is_group = 0
    is_text = 0
    is_endmark = 0
    weights = (0, 0, 0) # depth, length, lineno


class Single(Texel):
    is_single = 1
    weights = (0, 1, 0)
    style = EMPTYSTYLE
    text = '-'

    def __init__(self, style=None):
        if style:
            self.style = style

    def set_style(self, style):
        clone = shallow_copy(self)
        clone.style = style
        return clone

    def __setstate__(self, state):
        self.__dict__ = state.copy()
        self.style = as_style(self.style)


class Text(Texel):
    is_text = 1
    def __init__(self, text, style=EMPTYSTYLE):
        self.text = text
        self.style = style
        self.weights = (0, len(text), 0)

    def __repr__(self):
        return "T(%s)" % repr(self.text)

    def __setstate__(self, state):
        self.__dict__ = state.copy()
        self.style = as_style(self.style)

T = Text


class _TexelWithChilds(Texel):
    def compute_weights(self):
        if len(self.childs): # for empty groups, we use the default weights
            w_list = zip(*[child.weights for child in self.childs])
            self.weights = [f(l) for (l, f) in zip(w_list, self.functions)]



class Group(_TexelWithChilds):
    is_group = 1
    functions = (
        lambda l:max(l)+1, 
        sum,               
        sum)

    def __init__(self, childs):
        # NOTE: childs are assumed to be lists and not to change!
        # Since violations of these requirements can lead to difficult
        # to find bugs, we are making a copy of the list here,
        # although strictly this should not be necessary.
        self.childs = list(childs[:])
        self.compute_weights()

    def __repr__(self):
        return 'G(%s)' % repr(list(self.childs))
G = Group



class Container(_TexelWithChilds):
    is_container = 1
    functions = (
        lambda l:0,
        sum,               
        sum)
        
    def set_childs(self, childs):
        clone = shallow_copy(self)
        clone.childs = list(childs[:])
        clone.compute_weights()
        return clone

    def __repr__(self):
        return 'C(%s)' % repr(list(self.childs))

    def get_mutability(self):
        return [i%2==1 for i in range(len(self.childs))]

        

class NewLine(Single):
    weights = (0, 1, 1)
    style = EMPTYSTYLE
    parstyle = EMPTYSTYLE
    text = u'\n'

    def __repr__(self):
        return 'NL'

    def set_parstyle(self, style):
        clone = shallow_copy(self)
        clone.parstyle = style
        return clone

    def __setstate__(self, state):
        self.__dict__ = state.copy()
        self.style = as_style(self.style)
        self.parstyle = as_style(self.parstyle)


class Tabulator(Single):
    text = u'\t'

    def __repr__(self):
        return 'TAB'


class Fraction(Container):
    # A simple math object for debugging
    def __init__(self, denominator, nominator):
        self.childs = [TAB, denominator, TAB, nominator, TAB]
        self.compute_weights()
    

TAB = Tabulator()
NL = NewLine()
ENDMARK = NewLine()
ENDMARK.is_endmark = 1
NULL_TEXEL = T(u'')

# ---- functions -----
def depth(texel):
    """Returns the depth of an element.

       The depth can be definied as the number of generations of
       groups. By definition empty groups have depth 0.

       >>> depth(Text('abc'))
       0

       >>> depth(G(Text('abc')))
       1

       >>> depth(G()) # empty groups are ignored
       0
    """
    return texel.weights[0]


def length(texel):
    """Returns the length of *texel* in index units."""
    return texel.weights[1]


def spans(texel):
    r = []
    mutable = texel.get_mutability()
    k = -1    
    for i1, i2, child in iter_childs(texel):
        k += 1
        if mutable[k]:
            r.append((i1, i2))
    return r


def iter_childs(texel):
    assert texel.is_group or texel.is_container
    i1 = 0
    for child in texel.childs:
        i2 = i1+length(child)
        yield i1, i2, child
        i1 = i2    


def iter_d0(texel): # not needed, but might be useful in future
    """Iterate through all depth-zero elements. """
    l = [[texel]]
    i1 = 0
    while l:
        ll = l[-1]
        elem = ll[0]
        del ll[0]
        if elem.is_group:
            l.append(list(elem.childs))
        else:
            i2 = i1+length(elem)
            yield i1, i2, elem
            while l and not l[-1]:
                l.pop()


def iter_leaves(texel): # not needed, but might be useful in future
    """Iterate through all leaf-elements """
    l = [[texel]]
    i1 = 0
    while 1:
        while l and not l[-1]:
            l.pop()
        if not l:
            break
        ll = l[-1]
        elem = ll[0]
        del ll[0]
        if elem.is_group or elem.is_container:
            l.append(list(elem.childs))
        else:
            i2 = i1+length(elem)
            yield i1, i2, elem
            i1 = i2


def groups(l):
    """Transform the list of texels *l* into a list of groups.

       If texels have depth d, groups will have depth d+1. All
       returned groups are efficient.

       pre:
           is_elementlist(l)
           is_list_efficient(l)
           is_clean(l)

       post[l]:
           is_homogeneous(__return__)
           calc_length(l) == calc_length(__return__)
           #out("groups check: ok")
           #is_list_efficient(__return__) # XXX
           depth(__return__[0]) == depth(l[0])+1

    """
    r = []
    N = len(l)
    if N == 0:
        return r

    n = N // nmax
    if n*nmax < N:
        n = n+1
    rest = n*nmax-N

    a = nmax-float(rest)/n
    b = a
    while l:
        i = int(b+1e-10)
        r.append(Group(l[:i]))
        l = l[i:]
        b = b-i+a
    return r


def _join(l1, l2):
    """
    pre:
       is_homogeneous(l1)
       is_homogeneous(l2)
       is_clean(l1)
       is_clean(l2)
    post:
       is_clean(__return__)
       calc_length(l1)+calc_length(l2) == calc_length(__return__)
    """
    l1 = filter(length, l1) # strip off empty elements
    l2 = filter(length, l2) #
    if not l1:
        return l2
    if not l2:
        return l1
    t1 = l1[-1]
    t2 = l2[0]
    d1 = depth(t1)
    d2 = depth(t2)
    if d1 == d2:
        return l1+l2
    elif d1 > d2:
        return l1[:-1]+groups(_join(t1.childs, l2))
    # d1 < d2
    return groups(_join(l1, t2.childs))+l2[1:]


def join(*args):
    """Join several homogeneous lists of elements.

       The returned list is homogeneous.

       pre:
           forall(args, is_elementlist)
           forall(args, is_list_efficient)
       post[args]:
           is_homogeneous(__return__)
           is_clean(__return__)
           sum([calc_length(x) for x in args]) == calc_length(__return__)
           #out("join check: ok")
    """
    return reduce(_join, args)


def _fuse(l1, l2):
    """
    pre:
       is_homogeneous(l1)
       is_homogeneous(l2)
       is_clean(l1)
       is_clean(l2)
    post:
       is_clean(__return__)
       is_homogeneous(l2)
       calc_length(l1)+calc_length(l2) == calc_length(__return__)
    """
    l1 = filter(length, l1) # strip off empty elements
    l2 = filter(length, l2) #
    if not l1:
        return l2
    if not l2:
        return l1
    t1 = l1[-1]
    t2 = l2[0]
    left = get_rightmost(t1)
    right = get_leftmost(t2)
    if can_merge(left, right):
        new = merge(left, right)
        l1.pop()
        l1.append(exchange_rightmost(t1, new))
        return join(l1, remove_leftmost(t2), l2[1:])
    return _join(l1, l2)
    

def fuse(*args):
    """Like join(...) but also merge the arguments if possible.

       The returned list is homogeneous.

       pre:
           forall(args, is_elementlist)
           forall(args, is_homogeneous)

       post[args]:
           is_homogeneous(__return__)
           is_clean(__return__)
           sum([calc_length(x) for x in args]) == calc_length(__return__)
           #out("join check: ok")
    """
    return reduce(_fuse, args)


def insert(texel, i, stuff):
    """Inserts the list *stuff* at position *i*.

       *Texel* must be root-efficient, *stuff* must be
       list-efficient. Note that insert can increase the texels
       depth. The returned list is always list efficient.

       pre:
           isinstance(texel, Texel)
           is_elementlist(stuff)
           is_list_efficient(stuff)

       post:
           calc_length(__return__) == length(texel)+calc_length(stuff)
           is_list_efficient(__return__)
           is_clean(__return__)
    """
    if not 0 <= i <= length(texel):
        raise IndexError(i)
    if texel.is_group:
        k = -1
        for i1, i2, child in iter_childs(texel):
            k += 1
            if i1 <= i <= i2:
                l = insert(child, i-i1, stuff)
                r1 = texel.childs[:k]
                r2 = texel.childs[k+1:]
                return join(r1, l, r2)
    elif texel.is_container:
        mutable = texel.get_mutability()
        k = -1
        for i1, i2, child in iter_childs(texel):
            k += 1
            if (i1 < i < i2) or ((i1 <= i <= i2) and mutable[k]):
                l = insert(child, i-i1, stuff)
                r1 = texel.childs[:k]
                r2 = texel.childs[k+1:]
                return [texel.set_childs(r1+[grouped(l)]+r2)]
        if i == 0:
            return join(stuff, [texel])
        elif i == length(texel):
            return join([texel], stuff)
        assert False

    return fuse(copy(texel, 0, i), stuff, copy(texel, i, length(texel)))


def takeout(texel, i1, i2):
    """Takes out all content between *i1* and *i2*.

    Returns the rest and the cut out piece (kernel), i.e.
    G([a, b, c]).takeout(i1, i2) will return G([a, c]), b.

    *Texel* must be root efficient. Kernel and rest are guaranteed to
    be list efficient. Depths can change.

    pre:
        is_root_efficient(texel)
        #out(texel, i1, i2)

    post:
        is_elementlist(__return__[0])
        is_elementlist(__return__[1])
        is_homogeneous(__return__[0])
        is_homogeneous(__return__[1])
        calc_length(__return__[0])+i2-i1 == length(texel)
        calc_length(__return__[1]) == i2-i1
        #out("takeout", texel, i1, i2)
        #out(__return__[0])
        #out(__return__[1])
        #dump_list(__return__[0])
        is_clean(__return__[0])
        is_clean(__return__[1])
        is_list_efficient(__return__[0])
        is_list_efficient(__return__[1])

    """
    if not (0 <= i1 <= i2 <= length(texel)): 
        raise IndexError([i1, i2])
    if not length(texel): # 1. empty texel
        return [], []
    if i1 == i2:          # 2. empty interval
        return strip2list(texel), []
    if i1 <= 0 and i2 >= length(texel): # 3. fully contained
        return [], strip2list(texel)

    # Note that singles always fall under case 2 or 3. Beyond this
    # point we only have G, C or T.

    if texel.is_group:
        r1 = []; r2 = []; r3 = []; r4 = []
        k1 = []; k2 = []; k3 = []
        for j1, j2, child in iter_childs(texel):
            if j2 <= i1: #1
                r1.append(child)
            elif i1 <= j1 <= j2 <= i2: #2
                k2.append(child)
            elif j1 <= i1 <= j2: #3
                r, k = takeout(child, max(i1-j1, 0), min(i2-j1, length(child)))
                r2.extend(r)
                k1.extend(k)
            elif j1 <= i2 <= j2: #4
                r, k = takeout(child, max(i1-j1, 0), min(i2-j1, length(child)))
                r3.extend(r)
                k3.extend(k)
            elif i2 <= j1: #5
                r4.append(child)
        # Note that we are returning a list of elements which have
        # been in the content before. So even if texel is only root
        # efficient, the elements muss be element efficient.  Each of
        # the list r1, r2, r3, r4 and k1, k2, k3 is
        # homogeneous. Therefore join gives us list efficient return
        # values.
        if not is_clean(r2):
            dump_list(r2)
        if not is_clean(r3):
            dump_list(r3)

        tmp = fuse(r2, r3)
        return join(r1, tmp, r4), join(k1, k2, k3)
        
    elif texel.is_container:
        for k, (j1, j2, child) in enumerate(iter_childs(texel)):
            if  i1 < j2 and j1 < i2: # test of overlap
                if not (j1 <= i1 and i2 <= j2):
                    raise IndexError((i1, i2))
                childs = list(texel.childs) # this always creates a new list!
                tmp, kernel = takeout(
                    child, max(0, i1-j1), min(length(texel), i2-j1))
                childs[k] = grouped(tmp)
                return [texel.set_childs(childs)], kernel
        raise IndexError((i1, i2))

    elif texel.is_text:
        r1 = texel.text[:i1]
        r2 = texel.text[i2:]
        r3 = texel.text[i1:i2]
        s = texel.style
        return [Text(r1+r2, s)], [Text(r3, s)]

    assert False


def copy(root, i1, i2):
    """Copy all content of *root* between *i1* and *i2*.

       pre:
           isinstance(root, Texel)

       post:
           calc_length(__return__) == (i2-i1)
    """
    return takeout(root, i1, i2)[1]


def grouped(stuff):
    """Creates a single group from the list of texels *stuff*.

       If the number of texels exceeds nmax, subgroups are formed.
       Therefore, the depth can increase. 

       pre:
           is_elementlist(stuff)
           is_homogeneous(stuff)
           is_clean(stuff) 
           
       post:
           length(__return__) == calc_length(stuff)
    """
    while len(stuff) > nmax:
        stuff = groups(stuff)
    g = Group(stuff)
    return strip(g)


def strip(element):
    """Removes unnecessary Group-elements from the root."""
    n = length(element)
    while element.is_group and len(element.childs) == 1:
        element = element.childs[0]
    assert n == length(element)
    return element


def strip2list(texel):
    """Returns a list of texels which is list efficient. 
       pre:
           is_root_efficient(texel)
       post:
           is_list_efficient(__return__)
    """
    if texel.is_group:
        return texel.childs
    return [texel]


def provides_childs(texel):
    """Returns True if *texel* provides the childs_interface."""
    return texel.is_group or texel.is_container


def get_rightmost(element):
    if provides_childs(element):
        return get_rightmost(element.childs[-1])
    return element


def get_leftmost(element):
    if provides_childs(element):
        return get_leftmost(element.childs[0])
    return element


def exchange_rightmost(element, new):
    """Replace the rightmost subelement of *element* by *new*.

       pre:
           depth(new) == 0
       post:
           depth(__return__) == depth(element)
    """
    if provides_childs(element):
        l = exchange_rightmost(element.childs[-1], new)
        return Group(element.childs[:-1]+[l])
    return new


def remove_leftmost(element):
    """Removes the leftmost subelement of *element*. Returns a list.

       Note that this function can change the depth. 

       post:
           is_homogeneous(__return__)
           is_list_efficient(__return__)
    """

    if length(element) == 0 or depth(element) == 0:
        return []

    l = remove_leftmost(element.childs[0])
    return join(l, element.childs[1:])


def can_merge(texel1, texel2):
    return texel1.is_text and texel2.is_text and \
           texel1.style is texel2.style


def merge(texel1, texel2):
    """Merge the rightmost child of *texel1* with the leftmost child of
*texel2*.

       pre:
           isinstance(texel1, Texel)
           isinstance(texel2, Texel)
           can_merge(texel1, texel2)

       post:
           length(__return__) == length(texel1)+length(texel2)
    """
    return Text(texel1.text+texel2.text, texel1.style)


def compute_hull(texel, i1, i2, i0=0):
    if texel.is_text or texel.is_single:
        return i1, i2
    elif texel.is_container:
        overlapp = False
        for j1, j2 in spans(texel):
            if j1+i0 <= i1 <= i2 <= j2+i0: # inside
                overlapp = True
                continue
            if i1 < j2+i0 and j1+i0 < i2: # overlapp but not inside
                overlapp = True
                i1 = min(i1, j1+i0)
                i2 = max(i2, j2+i0)
        if not overlapp:
            i1 = min(i0, i1)
            i2 = max(i0+length(texel), i1)

    for j1, j2, child in iter_childs(texel):
        if j1+i0 <= i1 <= i2 <= j2+i0:
            i1, i2 = compute_hull(child, i1, i2, i0+j1)
    return i1, i2


def get_text(texel):
    if texel.is_single or texel.is_text:
        return texel.text
    assert texel.is_group or texel.is_container
    return u''.join([get_text(x) for x in texel.childs])


# ---- Debug Tools ---

def get_pieces(texel):
    """For debugging: returns a list of text pieces in *texel*."""
    if type(texel) == list:
        texel = grouped(texel)
    if provides_childs(texel):
        r = []
        for child in texel.childs:
            r.extend(get_pieces(child))
        return r
    if texel.is_text:
        return [texel.text]
    return [' '] # Glyph


def calc_length(l):
    """Calculates the total length of all elements in list *l*."""
    return sum([length(x) for x in l])


def is_clean(l):
    """Returns True if no element in list *l* has zero length. """
    for texel in l:                
        if length(texel) == 0:
            return False
        if texel.is_group:
            if not is_clean(texel.childs):
                return False
    return True
    

def is_homogeneous(l):
    """Returns True if the elements in list *l* are homogeneous.

       Elements are homogeneous if they have alle the same depth.
    """
    return maxdepth(l) == mindepth(l)


def is_efficient(element):
    """Returns True if *element* is efficient.

       An element is efficient, if it is not a group or otherwise if
       all of the following criteria are fulfilled:

       1. if each descendant group has between nmax/2 and nmax childs
       2. the depth of all childs is homogeneous

       In an efficient tree all groups (with the exception of the root
       node) must be element_efficient.

       Note: this function is very slow and should only be used for
       debugging.
    """
    if not is_group(element):
        return True
    if not is_homogeneous(element.childs):
        return False
    if len(element.childs) < nmax/2 or len(element.childs) > nmax:
        return False
    if not is_list_efficient(element.childs):
        return False
    return True


def is_list_efficient(l):
    """Returns True if each element in the list *l* is efficient.
       pre:
           isinstance(l, list)
    """
    if not is_homogeneous(l):
        return False
    for element in l:
        if not is_efficient(element):
            return False
    return True


def is_root_efficient(root):
    """Returns True if *root* is efficient.

       The tree spawned by *root* is efficient if it is either not a
       group or otherwise if it fulfills all of the following
       conditions:

       1. all childs are efficient
       2. the number of childs is <= nmax

       Note: this function is very slow and should only be used for
       debugging.
    """
    if not is_group(root):
        return True
    if len(root.childs) > nmax:
        return False
    return is_list_efficient(root.childs)


def is_group(element):
    return element.is_group


def is_elementlist(l):
    """Returns True if *l* is a list of Elements.
    """
    if not type(l) in (tuple, list):
        print "not a list or tuple", type(l), l.__class__ # XXX remove this
        return False
    return not False in [isinstance(x, Texel) for x in l]


def maxdepth(l):
    """Computes the maximal depth of all elements in list *l*.
    """
    m = [depth(x) for x in l if length(x)]
    if not m:
        return 0
    return max(m)


def mindepth(l):
    """Computes the minimal depth of all elements in list *l*.
    """
    m = [depth(x) for x in l if length(x)]
    if not m:
        return 0
    return min(m)


def out(*args):
    print repr(args)[1:-1]
    return True


def dump(texel, i=0):
    print (" "*i)+str(texel.__class__.__name__), texel.weights,
    if texel.is_text:
        print repr(texel.text), texel.style
    elif isinstance(texel, NewLine):
        print texel.parstyle
    else:
        print
    if texel.is_group or texel.is_container:
        for i1, i2, child in iter_childs(texel):
            dump(child, i+2)
    return True


def dump_list(l):
    print "Dumping list (efficient: %s)" % is_list_efficient(l)
    for i, element in enumerate(l):
        print "Dumping element no. %i" % i,
        print "(efficient: %s)" % is_efficient(element)
        dump(element)
    return True


# ---- Testing ----

if debug: # enable contract checking
     import contract
     contract.checkmod(__name__)


def test_00():
    "insert"
    set_nmax(4)
    s = G([T('1'), T('2'), T('3')])
    element = G([s, s, s])

    tmp = grouped(insert(element, 3, [T('X'), T('Y')]))
    assert depth(tmp) == depth(element)
    assert get_pieces(tmp) == ['1', '2', '3X', 'Y', '1', '2', '3', '1', '2', '3']

    tmp = insert(element, 3, [G([T('X'), T('Y')])])
    assert depth(grouped(tmp)) == depth(element)
    assert get_pieces(tmp) == ['1', '2', '3X', 'Y', '1', '2', '3', '1', '2', '3']


def test_01():
    "growth in insert"
    set_nmax(3)
    g = Group([T("a"), T("b"), T("c")])
    assert depth(g) == 1
    assert get_pieces(g) == ['a', 'b', 'c']
    n = grouped(insert(g, 0, [T("d")]))
    assert get_pieces(n) == ['da', 'b', 'c']


def test_02():
    "maintaining tree efficency in insert / takeout"
    set_nmax(3)
    texel = Group([])
    import random
    for i in range(100):
        x = random.choice("abcdefghijklmnopqrstuvwxyz")
        tmp = insert(texel, i, [Text(x)])
        assert is_list_efficient(tmp)
        texel = grouped(tmp)
    assert is_root_efficient(texel)
    #dump(texel)

    while length(texel):
        i1 = random.randrange(length(texel)+1)
        i2 = random.randrange(length(texel)+1)
        i1, i2 = sorted([i1, i2])
        r, k = takeout(texel, i1, i2)
        assert is_list_efficient(r)
        texel = grouped(r)
    assert is_root_efficient(texel)


def test_03():
    "depth"
    assert depth(Text('abc')) == 0
    assert depth(G([Text('abc')])) == 1
    assert depth(G([])) == 0
    element = G([G([G([T('1'), T('2')]), G([T('3')])])])
    assert depth(element) == 3


def test_04():
    "fuse"
    l1 = [T("01234")]
    l2 = [T("56789")]
    t = fuse([G(l1)], l2)
    assert get_pieces(t) == ['0123456789']

    t = fuse([G(l1)], [G(l2)])
    assert get_pieces(t) == ['0123456789']

    t = fuse(l1, [G(l2)])
    assert get_pieces(t) == ['0123456789']


def test_05():
    "iter_d0"
    t1 = T("012345678")
    t2 = T("ABC")
    t3 = T("xyz")
    c = Container().set_childs((t2, t3))
    g = G((t1, c))
    #texeltree.dump(g)
    l = []
    for i1, i2, elem in iter_d0(g):
        l.append((i1, i2, elem))
    assert repr(l) == "[(0, 9, T('012345678')), (0, 6, C([T('ABC'), T('xyz')]))]"
    

def test_06():
    "insert in container"
    fraction = Fraction(T("Sin(alpha)"), T("Cos(alpha)"))
    t = T("X") #, dict(color='red'))
    l = insert(fraction, 0, [t])
    assert get_pieces(grouped(l)) == ['X', ' ', 'Sin(alpha)', ' ', 'Cos(alpha)', ' ']
    l = insert(fraction, 1, [t])
    assert get_pieces(grouped(l)) == [' ', 'XSin(alpha)', ' ', 'Cos(alpha)', ' ']
    l = insert(fraction, 2, [t])
    assert get_pieces(grouped(l)) == [' ', 'SXin(alpha)', ' ', 'Cos(alpha)', ' ']
    l = insert(fraction, 11, [t])
    assert get_pieces(grouped(l)) == [' ', 'Sin(alpha)X', ' ', 'Cos(alpha)', ' ']
    l = insert(fraction, 12, [t])
    assert get_pieces(grouped(l)) == [' ', 'Sin(alpha)', ' ', 'XCos(alpha)', ' ']
    l = insert(fraction, 22, [t])
    assert get_pieces(grouped(l)) == [' ', 'Sin(alpha)', ' ', 'Cos(alpha)X', ' ']
    l = insert(fraction, 23, [t])
    assert get_pieces(grouped(l)) == [' ', 'Sin(alpha)', ' ', 'Cos(alpha)', ' ', 'X']

    elem = fraction
    text = "0123456789"
    for i in range(len(text)):
        tmp = insert(elem, 12, [T(text[i])])
        elem = grouped(tmp)
    assert get_pieces(elem) == [' ', 'Sin(alpha)', ' ', '9876543210Cos(alpha)', ' ']

    fraction = Fraction(T(""), T("Cos(alpha)"))
    l = insert(fraction, 0, [t])
    assert get_pieces(grouped(l)) == ['X', ' ', '', ' ', 'Cos(alpha)', ' ']

    l = insert(fraction, 1, [t])
    assert get_pieces(grouped(l)) == [' ', 'X', ' ', 'Cos(alpha)', ' ']

    l = insert(fraction, 2, [t])
    assert get_pieces(grouped(l)) == [' ', '', ' ', 'XCos(alpha)', ' ']


def test_07():
    "compute_hull"
    fraction = Fraction(T("Sin(alpha)"), T("Cos(alpha)"))
    n = length(fraction)
    assert compute_hull(fraction, 0, 1) == (0, n)
    assert compute_hull(fraction, 1, 2) == (1, 2)
    assert compute_hull(fraction, 2, 3) == (2, 3)
    assert compute_hull(fraction, 9, 11) == (9, 11)
    assert compute_hull(fraction, 11, 12) == (0, n)
    assert compute_hull(fraction, n-2, n-1) == (n-2, n-1)
    assert compute_hull(fraction, n-1, n) == (0, n)


def test_08():
    "takeout"
    s1 = as_style(dict(color='red'))
    s2 = as_style(dict(color='blue'))
    t1 = T("0123456789", style=s1)
    t2 = T("ABC", style=s2)
    g = G((t1, t2))
    r, k = takeout(g, 5, 12)    
    assert get_text(grouped(r)) == "01234C"
    assert get_text(grouped(k)) == "56789AB"

    t1 = T("0123456789")
    t2 = T("ABC")
    g = G((t1, t2))    
    r, k = takeout(g, 5, 12)
    assert get_pieces(grouped(r)) == ['01234C']
    assert get_text(grouped(k)) == "56789AB"

    t1 = T("0123456789")
    t2 = T("ABCDEFGHIJ")
    g = G((t1, t2))
    t = get_text(g)
    for i in range(20):
        for j in range(i, 20):
            r, k = takeout(g, i, j)
            assert get_text(grouped(r)) == t[:i]+t[j:]
            assert get_text(grouped(k)) == t[i:j]

    t1 = T("0123456789")
    t2 = T("ABCDEFGHIJ")
    t3 = T("klmnopqrst")
    t4 = T("UVWXYZ,.-*")
    g = G((t1, t2, t3, t4))
    t = get_text(g)
    for i in range(len(t)):
        for j in range(i, len(t)):
            r, k = takeout(g, i, j)
            assert get_text(grouped(r)) == t[:i]+t[j:]
            assert get_text(grouped(k)) == t[i:j]

    
def test_09():
    "pickle"
    s1 = as_style(dict(color='red'))
    s2 = as_style(dict(color='blue'))
    t1 = T("012345678", style=s1)
    t2 = T("ABC", style=s2)
    g = G((t1, t2))
    from cPickle import dumps, loads 
    t1_ = loads(dumps(t1))
    assert t1.style is t1_.style
    assert t1.text == t1_.text    
    g_ = loads(dumps(g))
    t1_, t2_ = g_.childs
    assert not t1_ is t1
    assert not t2_ is t2
    assert t1.style is t1_.style
    assert t2.style is t2_.style
    assert t1.text == t1_.text    
    assert t2.text == t2_.text    



# -*- coding: latin-1 -*-


from .textmodel import texeltree
from .textmodel.textmodel import TextModel
from .textmodel.texeltree import NewLine, Group, Text, length
from .textmodel.styles import EMPTYSTYLE
from .wxtextview.testdevice import TESTDEVICE
from .wxtextview.builder import BuilderBase
from .wxtextview.boxes import TextBox, NewlineBox, TabulatorBox, EmptyTextBox, \
    EndBox, check_box, Box, calc_length



class Builder(BuilderBase):
    parstyle = EMPTYSTYLE

    def __init__(self, device=TESTDEVICE):
        self.device = device
        self.model = model
        self._maxw = maxw

        import weakref
        self.cache = weakref.WeakValueDictionary()
        BuilderBase.__init__(self, model, device)

    def get_device(self):
        return self.device

    def mk_style(self, style):
        # This can overriden e.g. to implement style sheets. The
        # default behaviour is to use the paragraph style and add the
        # text styles.
        r = self.parstyle.copy()
        r.update(style)
        return r

    def extended_texel(self):
        return self.model.get_xtexel()
        
    def set_maxw(self, maxw):
        if maxw != self._maxw:
            self._maxw = maxw
            self.rebuild()

    ### Factory methods
    def create_boxes(self, texel):
        try:
            return self.cache[texel]
        except KeyError:
            pass

        name = texel.__class__.__name__+'_handler'
        handler = getattr(self, name)
        #print "calling handler", name
        l = handler(texel)
        try:
            assert calc_length(l) == length(texel)
        except:
            print "handler=", handler
            raise
        r = tuple(l)
        self.cache[texel] = r
        return r

    def Group_handler(self, texel):
        # Handles group texels. Note that the list of childs is
        # traversed from right to left. This way the "Newline" which
        # ends a line is handled before the content in the line. This
        # is important because in order to build boxes for the line
        # elements, we need the paragraph style which is located in
        # the NewLine-Texel.
        create_boxes = self.create_boxes
        r = ()
        for j1, j2, child in reversed(list(texeltree.iter_childs(texel))):
            r = create_boxes(child)+r
        return r

    # XXX Remove this:
    _cache = dict()
    _cache_keys = []
    def Text_handler(self, texel):
        # cached version
        key = texel.text, id(texel.style), id(self.parstyle), self.device
        try:
            return self._cache[key]
        except: pass        
        r = [TextBox(texel.text, self.mk_style(texel.style), 
                     self.device)]
        self._cache_keys.insert(0, key)        
        if len(self._cache_keys) > 10000:
            _key = self._cache_keys.pop()
            del self._cache[_key]
        self._cache[key] = r
        return r

    def Text_handler(self, texel):
        # non caching version
        return [TextBox(texel.text, self.mk_style(texel.style), 
                self.device)]

    def NewLine_handler(self, texel):
        self.parstyle = texel.parstyle
        if texel.is_endmark:
            return [self.EndBox(self.mk_style(texel.style), self.device)]
        return [NewlineBox(self.mk_style(texel.style), self.device)] # XXX: Hmmmm

    def Tabulator_handler(self, texel):
        return [TabulatorBox(self.mk_style(texel.style), self.device)]

    def TextCell_handler(self, texel):
        textbox = self.create_parstack(Group(texel.childs[1:]))
        cell = TextCellBox(textbox, device=self.device)
        assert len(cell) == length(texel)
        return [cell]

    def ScriptingCell_handler(self, texel):
        #dump_range(texel, 0, length(texel))
        
        # XXX TODO: implement proper treatment of temp
        sep1, (j1, j2, inp), sep2, (k1, k2, outp), sep3 \
            = texeltree.iter_childs(texel)

        _inp = Group(cell.childs[1:3])
        client = self._clients.get_matching(cell)
        colorized = client.colorize(_inp)
        inbox = self.create_parstack(colorized)
        assert len(inbox) == length(_inp)
        
        outbox = self.create_parstack(_outp)
        cell = ScriptingCellBox(inbox, outbox, number=texel.number,
                                device=self.device)
        assert len(cell) == n
        return [cell]

    def Plot_handler(self, texel):
        return [PlotBox(device=self.device)]

    def Graphics_handler(self, texel):
        return [GraphicsBox(texel, device=self.device)]

    def BitmapRGB_handler(self, texel):
        w, h = texel.size
        bitmap = wx.BitmapFromBuffer(w, h, texel.data)
        return [BitmapBox(bitmap, device=self.device)]

    def BitmapRGBA_handler(self, texel):
        w, h = texel.size
        im = wx.ImageFromData(w, h, texel.data)
        im.SetAlphaBuffer(texel.alpha)
        bitmap = wx.BitmapFromImage(im)
        return [BitmapBox(bitmap, device=self.device)]

    ### Builder methods
    def create_paragraphs(self, texel):
        # Ich müsste jeden Paragraphen separat erzeugen!
        boxes = self.create_boxes(texel)
        if self._maxw:
            maxw = max(100, self._maxw-80)
        else:
            maxw = 0
        l = create_paragraphs(
            boxes, maxw = maxw,
            Paragraph = self.Paragraph,
            device = self.device)
        return l

    def create_parstack(self, texel):
        l = self.create_paragraphs(texel, 0, length(texel))
        return ParagraphStack(l, device=self.device)

    def rebuild(self):
        model = self.model
        boxes = self.create_boxes(model.texel)
        self._layout = VGroup(boxes, device=self.device)


    ### Signal handlers
    def properties_changed(self, i1, i2):
        self.rebuild()

    def inserted(self, i, n):
        self.rebuild()

    def removed(self, i, n):
        #print "removed", i, n
        self.rebuild()





def test_00():
    model = TextModel(u'for a in range(5):\n    print a')
    cell = ScriptingCell(tmp.texel, NULL_TEXEL)
    model.insert(len(model), mk_textmodel(cell))

    assert find_cell(model.texel, 1) == (0, cell)

    view = ns['view']
    view.index = 1
    layout = view.builder.get_layout()
    r1 = layout.get_rect(0, 0, 0)
    assert r1.x2-r1.x1 == sepwidth

    r2 = layout.get_cursorrect(0, 0, 0, {})
    r3 = layout.get_cursorrect(len(model), 0, 0, {})
    assert r3.x2-r3.x1 == sepwidth
    assert r3.y1 > r2.y1



# -*- coding: latin-1 -*-


import wx
import cPickle
import string

from ..textmodel import TextModel
from ..textmodel.styles import updated_style
from .textview import TextView
from .wxdevice import WxDevice, defaultstyle, DCStyler
from .testdevice import TESTDEVICE
from .simplelayout import Builder

from math import ceil





class WXTextView(wx.ScrolledWindow, TextView):
    _scrollrate = 10, 10

    def __init__(self, parent, id=-1,  
                 pos=wx.DefaultPosition, size=wx.DefaultSize, style=0):
        wx.ScrolledWindow.__init__(self, parent, id,
                                   pos, size,
                                   style|wx.WANTS_CHARS)
        TextView.__init__(self)
        try:
            # This attribute only exists in some newer versions of wx (>= 3.0 ?) 
            wx.ScrolledWindow.DisableKeyboardScrolling(self)
        except AttributeError: pass
        self.Bind(wx.EVT_PAINT, self.on_paint)
        self.Bind(wx.EVT_ERASE_BACKGROUND, lambda event: None)
        self.Bind(wx.EVT_CHAR, self.on_char)
        self.Bind(wx.EVT_SIZE, self.on_size)
        self.Bind(wx.EVT_LEFT_DOWN, self.on_leftdown)
        self.Bind(wx.EVT_LEFT_DCLICK, self.on_leftdclick)
        self.Bind(wx.EVT_MOTION, self.on_motion)
        self.Bind(wx.EVT_KILL_FOCUS, self.on_focus)
        self.Bind(wx.EVT_SET_FOCUS, self.on_focus)
        
        # key = keycode, control, alt
        self.actions = {
            (wx.WXK_ESCAPE, False, False) : 'dump_info', 
            (wx.WXK_ESCAPE, True, False) : 'dump_boxes', 
            (wx.WXK_RIGHT, True, False)  : 'move_word_end',
            (wx.WXK_RIGHT, False, False)  : 'move_right',
            (wx.WXK_LEFT, True, False)  : 'move_word_begin',
            (wx.WXK_LEFT, False, False)  : 'move_left',
            (wx.WXK_DOWN, True, False)  : 'move_paragraph_end',
            (wx.WXK_DOWN, False, False)  : 'move_down',
            (wx.WXK_UP, True, False)  : 'move_paragraph_begin',
            (wx.WXK_UP, False, False)  : 'move_up',
            (wx.WXK_HOME, False, False) : 'move_line_start',
            (wx.WXK_END, False, False) : 'move_line_end',   
            (wx.WXK_HOME, True, False) : 'move_document_start',
            (wx.WXK_END, True, False) : 'move_document_end',            
            (wx.WXK_PAGEDOWN, False, False): 'move_page_down',
            (wx.WXK_PAGEUP, False, False): 'move_page_up',
            (wx.WXK_PAGEUP, True, False) : 'move_document_start',
            (wx.WXK_PAGEDOWN, True, False) : 'move_document_end',            
            (wx.WXK_RETURN, False, False): 'insert_newline',
            (wx.WXK_BACK, False, False): 'backspace',
            (wx.WXK_DELETE, False, False): 'delete',
            (3, True, False) : 'copy',
            (22, True, False) : 'paste',
            (24, True, False) : 'cut',
            (26, True, False) : 'undo',
            (18, True, False) : 'redo',  
            (11, True, False) : 'del_line_end',   
            (127, True, False) : 'del_word_left',   
            (1, True, False) : 'select_all',
            (9, True, False) : 'indent',
            (21, True, False) : 'dedent',
            }        
        
    def create_builder(self):
        return Builder(
            self.model, 
            device=WxDevice(), 
            maxw=self._maxw)

    def refresh(self):
        self.Refresh()

    def on_focus(self, event):
        # focus changed
        self.Refresh()

    def on_char(self, event):
        keycode = event.GetKeyCode()
        ctrl = event.ControlDown()
        shift = event.ShiftDown()
        alt = event.AltDown()        
        action = self.actions.get((keycode, ctrl, alt))
        if action is None:
            action = unichr(keycode)
        self.handle_action(action, shift)

    def to_clipboard(self, textmodel):
        text = textmodel.get_text()
        plain = wx.TextDataObject()
        plain.SetText(text)
        pickled = wx.CustomDataObject("pytextmodel")
        pickled.SetData(cPickle.dumps(textmodel))
        data = wx.DataObjectComposite()
        data.Add(plain)
        data.Add(pickled)
        wx.TheClipboard.Open()
        wx.TheClipboard.SetData(data)
        wx.TheClipboard.Close()

    def read_clipboard(self):
        if wx.TheClipboard.IsOpened():  # may crash, otherwise
            return
        pickled = wx.CustomDataObject("pytextmodel")
        plain = wx.TextDataObject()
        textmodel = None
        wx.TheClipboard.Open()
        if wx.TheClipboard.GetData(pickled):            
            textmodel = cPickle.loads(pickled.GetData())

        elif wx.TheClipboard.GetData(plain):
            textmodel = self._TextModel(plain.GetText())

        wx.TheClipboard.Close()
        return textmodel

    def on_paint(self, event):
        self._update_scroll()
        self.keep_cursor_on_screen()

        pdc = wx.PaintDC(self)
        pdc.SetAxisOrientation(True, False)
        if self.builder.get_device().buffering:
            dc = wx.BufferedDC(pdc)
            if not dc.IsOk():
                return
        else:
            dc = pdc
        dc.SetBackgroundMode(wx.SOLID)
        dc.SetBackground(wx.WHITE_BRUSH)
        dc.Clear()
        styler = DCStyler(dc)
        
        region = self.GetUpdateRegion()
        x, y, w, h = region.Box
        dc.SetClippingRegion(x-1, y-1, w+2, h+2)

        x, y = self.CalcScrolledPosition((0,0)) 
        layout = self.layout
        layout.draw(x, y, dc, styler)
        if wx.Window.FindFocus() is self: 
            layout.draw_cursor(self.index, x, y, dc, self.model.defaultstyle)
        for j1, j2 in self.get_selected():
            layout.draw_selection(j1, j2, x, y, dc)
        styler = None
        dc = None

    def on_size(self, event):
        self.keep_cursor_on_screen()

    def on_motion(self, event):
        if not event.LeftIsDown():
            return event.Skip()
        x, y = self.CalcUnscrolledPosition(event.Position) 
        i = self.layout.get_index(x, y)
        if i is not None:
            self.set_index(i, extend=True)

    def on_leftdown(self, event):
        x, y = self.CalcUnscrolledPosition(event.Position) 
        i = self.compute_index(x, y)
        if i is not None:
            self.set_index(i, extend=event.ShiftDown())
        self.SetFocus()

    def on_leftdclick(self, event):
        # Mark word
        x, y = self.CalcUnscrolledPosition(event.Position) 
        self.select_word(x, y)
        self.SetFocus()

    ### Scroll
    def _update_scroll(self):
        layout = self.layout
        self.SetVirtualSize((layout.width, layout.height))
        self._scrollrate = 10, 10
        self.SetScrollRate(*self._scrollrate)
        
    def adjust_viewport(self):
        # Adjust view so that the cursor is visible
        layout = self.layout
        r = layout.get_rect(self.index, 0, 0)
        fw, fh = self._scrollrate

        width, height = self.GetClientSize()
        firstcol, firstrow = self.GetViewStart() # -> Scroll in Inc-steps
        x = firstcol*fw
        y = firstrow*fh
        if r.y1 <= y: 
            # If r is below and above the viewport, we prefer to scroll up
            y = r.y1
            firstrow = int(y/fh)
        elif r.y2 > y+height:
            y = r.y2-height
            firstrow = ceil(y/float(fh))
        if r.x1 <= x:
            x = r.x1
            firstcol = int(x/fw)
        elif r.x2 > x+width:
            x = r.x2-width
            firstcol = ceil(x/float(fw))
        if (firstcol, firstrow) != self.GetViewStart():
            self.Update() # important update first!
            self.Scroll(firstcol, firstrow)

    def keep_cursor_on_screen(self):
        pass # not implemented
        
        
        

testtext = u"""Ein männlicher Briefmark erlebte
Was Schönes, bevor er klebte.
Er war von einer Prinzessin beleckt.
Da war die Liebe in ihm geweckt.
Er wollte sie wiederküssen,
Da hat er verreisen müssen.
So liebte er sie vergebens.
Das ist die Tragik des Lebens.

(Joachim Ringelnatz)"""

def init_testing(redirect=True):
    app = wx.App(redirect=redirect)
    model = TextModel(testtext)
    model.set_properties(15, 24, fontsize=14)
    model.set_properties(249, 269, fontsize=14)

    frame = wx.Frame(None)
    win = wx.Panel(frame)
    view = WXTextView(win)
    view.model = model
    assert view.layout is not None
    box = wx.BoxSizer(wx.VERTICAL)
    box.Add(view, 1, wx.ALL|wx.GROW, 1)
    win.SetSizer(box)
    win.SetAutoLayout(True)
    frame.Show()    
    return locals()


    
def test_02():
    ns = init_testing(redirect=True)
    view = ns['view']
    view.cursor = 5
    view.selection = 3, 6    
    return ns


def test_03():
    ns = init_testing(redirect=False)
    model = ns['model']
    view = ns['view']
    assert view.layout is not None

    model.set_properties(10, 20, fontsize=15)
    assert view.layout is not None

    n = len(model)
    text = '\n12345\n'
    model.insert_text(5, text)
    model.remove(5, 5+len(text))
    assert len(model) == n
    assert len(view.layout) == n+1
    return locals()


def test_04():
    "insert/remove"
    ns = init_testing(False)
    model = ns['model']
    view = ns['view']
    text = model.get_text()
    n = len(model)
    for i in range(len(text)):
        model.insert_text(i, 'X')
        assert len(model) == n+1
        model.remove(i, i+1)
        assert len(model) == n

def test_05():
    "remove"
    ns = init_testing(redirect=False)
    model = ns['model']
    view = ns['view']
    text = model.get_text()
    n = len(model)
    for i in range(len(text)-1):
        old = model.remove(i, i+1)
        assert len(model) == n-1
        model.insert(i, old)
        assert len(model) == n

def test_09():
    "linebreak"
    ns = init_testing(redirect=False)
    model = ns['model']
    view = ns['view']
    text = model.get_text()
    view.set_maxw(100)
    return ns

def test_10():
    ns = init_testing(redirect=False)
    model = ns['model']
    view = ns['view']
    model.remove(0, len(model))
    model.insert(0, TextModel("123\n"))

    builder = view.builder
    builder.set_maxw(100)
    layout = view.layout
    assert layout.get_info(4, 0, 0)
    x, y =  layout.get_info(3, 0, 0)[-2:]
    u, v = layout.get_info(4, 0, 0)[-2:]
    # Cursorposition 4 must be at the begining of next line
    assert u<x # 3 must be left of 4
    assert v>y # and 4 must be below 3


def test_11():
    # Problem: if the click position is right of the line, the cursor
    # should jump to the last index position in that line.
    ns = init_testing(redirect=False)
    model = ns['model']
    view = ns['view']
    #view.layout.dump_boxes(0, 0, 0)
    model.remove(0, len(model))
    model.insert(0, TextModel("123\n"))
    assert view.layout.get_index(100, 0) == 3


def test_13():
    # Problem: Exception during delete 10.01.2015
    ns = init_testing(redirect=False)
    model = ns['model']
    view = ns['view']
    view.index = 271
    view.selection = (42, 42+227)
    view.cut()
    
def test_14():
    "join_undo"
    ns = init_testing(redirect=False)
    view = ns['view']
    for i, text in enumerate('abcd'):
        view.add_undo(view.insert(i, TextModel(text)))
    assert len(view._undoinfo) == 1

    view._undoinfo = [] # reset undo

    # emulate backspace
    view.add_undo(view.remove(10, 11))
    view.add_undo(view.remove(9, 10))
    assert len(view._undoinfo) == 1

def demo_00():
    "simple demo"
    ns = test_02()
    import testing
    testing.pyshell(ns)    
    ns['app'].MainLoop()


def demo_01():
    "colorize demo"
    app = wx.App(redirect = False)
    frame = wx.Frame(None)
    win = wx.Panel(frame)
    view = WXTextView(win)
    box = wx.BoxSizer(wx.VERTICAL)
    box.Add(view, 1, wx.ALL|wx.GROW, 1)
    win.SetSizer(box)
    win.SetAutoLayout(True)

    from ..textmodel.textmodel import pycolorize
    from ..textmodel import texeltree
    filename = texeltree.__file__.replace('pyc', 'py')
    rawtext = open(filename).read() 
    model = pycolorize(rawtext)
    view.set_model(model)
    frame.Show()
    app.MainLoop()


def demo_02():
    "empty text"
    app = wx.App(redirect = True)
    frame = wx.Frame(None)
    win = wx.Panel(frame)
    view = WXTextView(win)
    box = wx.BoxSizer(wx.VERTICAL)
    box.Add(view, 1, wx.ALL|wx.GROW, 1)
    win.SetSizer(box)
    win.SetAutoLayout(True)
    model = TextModel(u'')
    view.set_model(model)
    frame.Show()
    import testing
    testing.pyshell(locals())    
    app.MainLoop()


def demo_03():
    "line break"
    ns = test_09()
    import testing
    testing.pyshell(ns)    
    ns['app'].MainLoop()

def benchmark_00():
    text = ""
    for i in range(1000):
        text += "Copy #%d \n" % i

    model = TextModel(u'Hello World!')
    model.set_properties(6, 11, fontsize=14)
    model.set_properties(0, 11, bgcolor='yellow')
    model.insert(len(model), TextModel(text))
    app = wx.App(False)

    frame = wx.Frame(None)
    view = WXTextView(frame, -1)
    view.model = model
    frame.Show()

    for i in range(100):
        model.insert_text(1000, "TEXT")

    
    
if __name__ == '__main__':
    from ..textmodel import alltests
    alltests.dotests()



from math import sqrt


class Rect:
    def __init__(self, x1, y1, x2, y2):
        self.x1 = min(x1, x2)
        self.y1 = min(y1, y2)
        self.x2 = max(x1, x2)
        self.y2 = max(y1, y2)
        
    def __repr__(self):
        return 'Rect%s' % repr((self.x1, self.y1, self.x2, self.y2))

    def dist(self, x, y):
        dx = max(self.x1 - x, 0, x - self.x2);
        dy = max(self.y1 - y, 0, y - self.y2)
        return sqrt(dx*dx + dy*dy)

    def adist(self, x, y):
        # alternative dist: puts stronger weight on y-deviations
        dx = max(self.x1 - x, 0, x - self.x2);
        dy = max(self.y1 - y, 0, y - self.y2)
        #print "adist:", self, x, y, dx, dy
        return dy*dy, dx*dx

    def items(self):
        return self.x1, self.y1, self.x2, self.y2

    def __eq__(self, other):
        return self.items() == other.items()


def combine_rects(r1, r2):
    if r1 is None:
        return r2
    elif r2 is None:
        return r1
    r = Rect(0, 0, 0, 0)
    r.x1 = min(r1.x1, r2.x1)
    r.x2 = max(r1.x2, r2.x2)
    r.y1 = min(r1.y1, r2.y1)
    r.y2 = max(r1.y2, r2.y2)
    return r



# -*- coding: latin-1 -*-

# cerealizerformat.py
#
# A simple file format based on cerealizer. This is basically a secure
# form of pickling. Is meant to be used during development.

from . import nbtexels
from .textmodel import textmodel
from .textmodel import texeltree, styles
from .nbtexels import mk_textmodel

from . import cerealizer
from StringIO import StringIO


magic = 'pynotebook0'

def register(Class, handler=None, classname = None):
    # By default we want to register a class by its short name,
    # however in cerealizer the default is the full name (module +
    # name). I prefer the short name, because classes may be moved to
    # other modules during development.
    if classname is None:
        classname = Class.__name__
    cerealizer.register(Class, handler, classname)



def dumps(obj):
    s = StringIO()
    cerealizer.Dumper(magic).dump(obj, s)
    return s.getvalue()


def _replace_styles(texel, table):
    if texel.is_group or texel.is_container:
        for child in texel.childs:
            _replace_styles(child, table)
    else:
        if 'style' in texel.__dict__:
            sid = id(texel.style)
            if not sid in table:
                table[sid] = styles.create_style(**texel.style)
            texel.style = table[sid]
        if isinstance(texel, texeltree.NewLine) and 'parstyle' in texel.__dict__:
            sid = id(texel.parstyle)
            if not sid in table:
                table[sid] = texel.parstyle = styles.create_style(**texel.parstyle)
            texel.parstyle = table[sid]

def loads(s):
    model =  cerealizer.Dumper(magic).undump(StringIO(s))
    _replace_styles(model.texel, {})
    return model


register(texeltree.Group)
register(texeltree.Text)
register(texeltree.Container)
register(texeltree.NewLine)
register(texeltree.Tabulator)

register(textmodel.TextModel)

register(nbtexels.TextCell)
register(nbtexels.ScriptingCell)
register(nbtexels.BitmapRGB)
register(nbtexels.BitmapRGBA)


def test_00():
    model1 = textmodel.TextModel()
    tmp = textmodel.TextModel(u'for a in range(5):\n    print a')
    cell = nbtexels.ScriptingCell(tmp.texel, texeltree.NULL_TEXEL)
    model1.insert(len(model1), mk_textmodel(cell))

    s = dumps(model1)
    model2 = loads(s)
    assert str(model1.texel) == str(model2.texel)
    assert model1.get_style(5) is model2.get_style(5)
    assert model1.get_parstyle(5) is model2.get_parstyle(5)
    try:
        model2 = loads('*'+s)
        assert False
    except cerealizer.NotCerealizerFileError:
        pass



# -*- coding: latin-1 -*-

"""
A simple notebook for the r language. Uses rpy2 for interfacing and 
pygments for colorization. 
"""

import sys
if __name__ == '__main__':
    sys.path.insert(0, '..')

from pynotebook.clients import Client
from pynotebook.pyclient import FakeFile
from pynotebook.nbstream import StreamRecorder
from pynotebook.nbtexels import ScriptingCell as _ScriptingCell
from pynotebook.nbview import TextModel, NBView
from pynotebook.textformat import fromtext
from pynotebook.textmodel import TextModel
from pynotebook.textmodel.texeltree import get_text, length

import wx

# we use rpy2 
import rpy2.robjects as robjects

# and pygments
from pygments.lexers import SLexer
from pygments.formatter import Formatter
from pygments import token as Token
from pygments import highlight

class TexelFormatter(Formatter):
    encoding = 'utf-8'
    def __init__(self, bgcolor):
        self.bgcolor = bgcolor

    def format(self, tokensource, outfile): 
        bgcolor = self.bgcolor
        self.model = model = TextModel()
        for token, text in tokensource:
            if token is Token.Keyword:
                style = dict(textcolor='red')
            elif token is Token.Literal.Number:
                style = dict(textcolor='blue')
            elif token is Token.Comment.Single:
                style = dict(textcolor='grey')
            elif token is Token.Text:
                style = dict(textcolor='black')
            elif token is Token.Literal.String:
                style = dict(textcolor='red')                    
            else:
                style = dict()
            style['bgcolor'] = bgcolor
            new = TextModel(text, **style)
            model.append(new)



class RClient(Client):
    name = "rpy2 r"
    def __init__(self):
        self.r = robjects.r

    def execute(self, inputfield, output):
        return self.run(get_text(inputfield), output)

    def run(self, code, output):
        self.counter += 1
        bkstdout, bkstderr = sys.stdout, sys.stderr
        sys.stdout = FakeFile(output)
        sys.stderr = FakeFile(lambda s: output(s, iserr=True))

        try:
            try:
                result = self.r(code)
            except ValueError, e:
                output(e, True)
            except Exception, e:
                output(e, True)
            else:
                output(result)
        except Exception, e:
            output(repr(e), True)
        finally:
            sys.stdout, sys.stderr = bkstdout, bkstderr
            sys.settrace(None)
    
    def complete(self, word, nmax=None):
        import rpy2.robjects 
        ri = rpy2.robjects.rinterface
        options = set()
        for env in (ri.baseenv, ri.globalenv): # XXX are there more envireonments??
            for name in env:
                if name.startswith(word):
                    options.add(name)
            if len(options) == nmax:
                break
        return options

    def colorize(self, inputtexel, styles=None, bgcolor='white'):
        text = get_text(inputtexel)
        assert len(text) == length(inputtexel)
        formatter = TexelFormatter(bgcolor)
        highlight(text, SLexer(), formatter)
        model = formatter.model
        while len(model) < length(inputtexel):
            # XXX when does this happen? What does it mean?
            model.insert_text(len(model), '\n')
        model = formatter.model[0:length(inputtexel)]
        assert len(model) == length(inputtexel)
        return model.texel




class ScriptingCell(_ScriptingCell):
    client_name = RClient.name



def test_00():
    "send"
    interpreter = RClient()

    stream = StreamRecorder()
    interpreter.run('1+2', stream.output)
    len(stream.messages) == 1
    stream.messages[-1] == False # no Error


examples = """[In 1]:
# Some R-Code

x <- c(1,2,3,4,5,6)   # Create ordered collection (vector)
y <- x^2              # Square the elements of x
print(y)              # print (vector) y
[In 0]:
# A simple plot 

plot(x, y)
[In 0]:
# A nice plot

numberWhite <- rhyper(30,4,5,3)
numberChipped <- rhyper(30,2,7,3)
smoothScatter(
    numberWhite,numberChipped,
    xlab="White Marbles",ylab="Chipped Marbles",
    main="Drawing Marbles")
"""

def demo_00():
    app = wx.App(redirect=False)
    model = fromtext(examples, ScriptingCell=ScriptingCell)

    frame = wx.Frame(None)
    win = wx.Panel(frame)
    view = NBView(win)
    view.ScriptingCell = ScriptingCell
    view._clients.register(RClient())
    view.model = model
    box = wx.BoxSizer(wx.VERTICAL)
    box.Add(view, 1, wx.ALL|wx.GROW, 1)
    win.SetSizer(box)
    win.SetAutoLayout(True)

    frame.Show()
    app.MainLoop()


if __name__ == '__main__':
    demo_00()




# -*- coding:utf-8 -*-

#
# graphics.py
#
#
# This module defines graphics operations which are used in connection
# with the 'Graphics'-Texel. Each instance of an operations definies a
# certain graphical object (e.g. a line or circle) or sets certain
# drawing properties (.e.g. line color or fill color).
#
# Lists of graphic operations represent a vector graphic. They can be
# saved, loaded or displayed.


# XXX missing: 
# - SetLogicalFunction
# - fill styles
# - gradient fills
# - rounded rectangle
# - clipping
# - spline


import wx
from .nbtexels import Graphics



def _normalize_color(color):
    if isinstance(color, wx.Colour):
        return color.asTuple()+(color.Alpha(), )
    elif isinstance(color, str):
        c = wx.Colour()
        c.SetFromString(color)
        return tuple(c,)+(c.Alpha(), )
    elif isinstance(color, list) or isinstance(color, tuple):
        return tuple(color)
    else:
        raise ValueError('Not a valid color specification: %s' % repr(color))


# Some versions of wx (accidentally) use degrees instead of
# radians. The following factor is used to correct this. It has the
# value None and is set later to either 1.0 or pi/180.0
#
# Calculation is could be done here in principle, using a
# MeasuringContext. However, this always gives memory violations.
_rotation_correction = None


    
class LineColor:
    def __init__(self, color):
        self.color = _normalize_color(color)
        
    def draw(self, gc, state):
        pen = state["pen"]
        pen.SetColour(self.color)
        gc.SetPen(pen)


class LineWidth:
    def __init__(self, width):
        self.width = width
        
    def draw(self, gc, state):
        pen = state["pen"]
        pen.SetWidth(self.width)
        gc.SetPen(pen)


class LineDashes:
    # Dashes is a list of length values. Example: [5, 5] will draw a
    # line of length 5, a space of length 5, a line of length 5, ...
    def __init__(self, dashes):
        self.dashes = dashes
        
    def draw(self, gc, state):
        pen = state["pen"]
        pen.SetDashes(self.dashes)
        gc.SetPen(pen)

line_joins = dict(
    bevel=wx.JOIN_BEVEL, miter=wx.JOIN_MITER, round=wx.JOIN_ROUND)

class LineJoin:
    def __init__(self, join_style):
        if not join_style in line_joins:
            raise ValueError("join_style mus be 'bevel, 'miter or 'round'")
        self.join_style = join_style
        
    def draw(self, gc, state):
        pen = state["pen"]
        pen.SetJoin(line_joins[self.join_style])
        gc.SetPen(pen)


line_caps = dict(
    round=wx.CAP_ROUND, projecting=wx.CAP_PROJECTING, butt=wx.CAP_BUTT)

class LineCap:
    def __init__(self, cap_style):
        if not cap_style in line_caps:
            raise ValueError("cap_style mus be 'round', 'projecting' or 'butt'")
        self.cap_style = cap_style
        
    def draw(self, gc, state):
        pen = state["pen"]
        pen.SetCap(line_caps[self.cap_style])
        gc.SetPen(pen)

        
class FillColor:
    def __init__(self, color):
        self.color = _normalize_color(color)
        
    def draw(self, gc, state):
        brush = state["brush"]
        brush.SetColour(self.color)
        brush.SetStyle(wx.SOLID)
        gc.SetBrush(brush) 


class Dot: # remove this?
    def __init__(self, (x, y)):
        self.x = x
        self.y = y

    def draw(self, gc, state):
        p = gc.CreatePath()
        p.AddCircle(self.x, self.y, 5)
        p.Transform(state['matrix'])
        gc.DrawPath(p)


class Line:
    def __init__(self, *points):
        self.points = tuple(points)

    def draw(self, gc, state):
        points = self.points
        if len(points)<2:
            return
        p = gc.CreatePath()
        for point in points:
            p.AddLineToPoint(point)
        p.Transform(state['matrix'])
        gc.DrawPath(p)


# class Spline:
#     def __init__(self, *points):
#         self.points = tuple(points)
#
#     def draw(self, gc, state):
#         gc.DrawSpline(self.points)
#         # XXX TODO


class Polygon:
    def __init__(self, *points):
        self.points = tuple(points)
        # XXX should we be able to set the fill rule?

    def draw(self, gc, state):
        points = self.points
        if len(points)<2:
            return
        p = gc.CreatePath()
        for point in points:
            p.AddLineToPoint(point)
        p.CloseSubpath()
        p.Transform(state['matrix'])
        gc.DrawPath(p)


        
class Path:
    """A path represents shapes like lines, curves, arcs and more.

*data* ist a list of drawing commands which are each specified as
 tuples with coordinate values. Currently there are four commands: 
  - M (move to point)
  - L (line to point)
  - Q (Bezier curve to point)
  - Z (close path)
    """
    def __init__(self, *data):
        self.data = data
        
    def draw(self, gc, state):
        wxpath = gc.CreatePath()
        x = y = 0
        for item in self.data:
            cmd = item[0]
            points = item[1:]
            if cmd == 'M':
                x, y = points
                wxpath.MoveToPoint(x, y)
            elif cmd == 'L':
                x, y = points
                wxpath.AddLineToPoint(x, y)
            elif cmd == 'Q':
                cx, cy, x, y = points
                wxpath.AddQuadCurveToPoint(cx, cy, x, y)
            elif cmd == 'Z':
                wxpath.CloseSubpath()
            else:
                raise Exception("Unknown path command: '%s'" % cmd)
        wxpath.Transform(state['matrix'])            
        gc.DrawPath(wxpath)



    
class Circle:
    def __init__(self, (x, y), r):
        self.x = x
        self.y = y
        self.r = r
        
    def draw(self, gc, state):
        p = gc.CreatePath()
        p.AddCircle(self.x, self.y, self.r)
        p.Transform(state['matrix'])
        gc.DrawPath(p)


class Ellipse:
    def __init__(self, (x, y), r1, r2):
        self.x = x
        self.y = y
        self.r1 = r1
        self.r2 = r2
        
    def draw(self, gc, state):
        p = gc.CreatePath()
        p.AddEllipse(self.x, self.y, self.r1, self.r2)
        p.Transform(state['matrix'])
        gc.DrawPath(p)

        

class Arc:
    def __init__(self, (x, y), r, start, end):
        self.x = x
        self.y = y
        self.r = r
        self.start = start
        self.end = end
        # XXX do we need clockwise?
        
    def draw(self, gc, state):
        p = gc.CreatePath()
        p.AddArc(self.x, self.y, self.r, self.start, self.end)
        p.Transform(state['matrix'])
        gc.DrawPath(p)



class Rectangle:
    def __init__(self, (x1, y1), (x2, y2)):
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2
        
    def draw(self, gc, state):
        x = min(self.x1, self.x2)
        w = max(self.x1, self.x2)-x
        y = min(self.y1, self.y2)
        h = max(self.y1, self.y2)-y        
        p = gc.CreatePath()
        p.AddRectangle(x, y, w, h)
        p.Transform(state['matrix'])
        gc.DrawPath(p)

        

class Font:
    def __init__(self, size, family, style, weight, 
                 underline=False, face="", 
                 encoding=wx.FONTENCODING_DEFAULT):
        self.size = size
        self.family = family
        self.style = style
        self.weight = weight
        self.underline = underline
        self.face = face
        self.encoding = encoding
        
    def draw(self, gc, state):
        font = wx.Font(self.size, self.family, self.style,
                       self.weight, self.underline, 
                       self.face, self.encoding)
        gc.SetFont(font)


class Text:
    regname = 'GraphicsText'
    def __init__(self, text, point, align=(0, 0)):
        self.text = text
        self.point = point
        self.align = align
        
    def draw(self, gc, state):
        x, y = self.point
        w, h = gc.GetTextExtent(self.text)
        dx = 0.5*w*(self.align[0]-1)
        dy = 0.5*h*(self.align[1]-1)
        gc.SetTransform(state['matrix'])
        gc.DrawText(self.text, x+dx, y+dy)
        gc.SetTransform(state['trafo'])


class Translate:
    def __init__(self, offset):
        self.offset = offset
        
    def draw(self, gc, state):
        dx, dy = self.offset
        state['matrix'].Translate(dx, dy) 
        #gc.Translate(dx, dy)


class Rotate:
    def __init__(self, angle):
        self.angle = angle
        
    def draw(self, gc, state):
        global _rotation_correction
        if _rotation_correction is None:
            # Determine rotation correction factor.
            m = gc.CreateMatrix()
            m.Rotate(360)
            if m.Get()[0]<-0.28:
                # radians
                _rotation_correction = 1
            else:
                # degrees
                import math
                _rotation_correction = math.pi/180
        else:
            _rotation_correction = 1

        state['matrix'].Rotate(self.angle*_rotation_correction) 


class Scale:
    def __init__(self, fx, fy=None):
        if fy is None:
            fy = fx
        self.fx = fx
        self.fy = fy
        
    def draw(self, gc, state):
        state['matrix'].Scale(self.fx, self.fy)



def register_classes():
    from cerealizerformat import register
    import types
    for name, value in globals().items():
        if type(value) is types.ClassType:
            if hasattr(value, 'regname'):
                register(value, classname=value.regname)
            else:
                register(value)


def init_testing(redirect=True):
    import wx
    from .nbview import TextModel, NBView
    app = wx.App(redirect=redirect)
    model = TextModel('')

    frame = wx.Frame(None)
    win = wx.Panel(frame, -1)
    view = NBView(win, -1, style=wx.SUNKEN_BORDER)
    view.model = model
    box = wx.BoxSizer(wx.VERTICAL)
    box.Add(view, 1, wx.ALL|wx.GROW, 1)
    win.SetSizer(box)
    win.SetAutoLayout(True)

    frame.Show()
    return locals()

def test_00():
    "Graphics"
    from .nbview import mk_textmodel
    ns = init_testing(False)
    model = ns['model']
    model.insert(len(model), mk_textmodel(Graphics([Circle((0, 0), 5)])))
    segments = [
        (0, 0), (0, 10), (10, 10), (10, 0)
    ]
    l = Line(*segments)
    model.insert(len(model), mk_textmodel(Graphics([l])))
    return ns


    
def demo_00():
    import wx
    from .wxtextview import testing
    ns = test_00()
    testing.pyshell(ns)
    py = ns['view']._clients._clients['direct python']
    py.namespace.update(
        dict(Graphics=Graphics, Dot=Dot, Line=Line))
    ns['app'].MainLoop()








# log method calls which lead to exceptions

import os
import sys
import traceback
import cPickle
from tempfile import mkdtemp

bk_stderr = sys.stderr

def store_exception(obj, name, args, kwds):
    dirname = mkdtemp(prefix="pnb_tmp")
    tb = traceback.format_exc()
    f = open(os.path.join(dirname, 'exception.txt'), 'wb')
    print >>f, tb
    f = open(os.path.join(dirname, 'call.pickle'), 'wb')
    l = (obj, name, args, kwds)
    cPickle.dump(l, f)
    print >>bk_stderr, tb
    print >>bk_stderr, "Stored details to file %s" % repr(dirname)


    
def controlled(f):
    def new_f(self, *args, **kwds):
        try:
            return f(self, *args, **kwds)
        except:
            store_exception(self.model, f.__name__, args, kwds)
            raise
    return new_f








# -*- coding: latin-1 -*-


from .texeltree import length, provides_childs, iter_childs, Texel


debug = 0


class NotFound(Exception): 
    pass


# NOTE: The following helper functions for searching indices only
# work with certain weight functions. They will work for weights
# aggregated by 'sum', such as lengths and line numbers. But trying to
# find depth values will lead to unexpected and unpredicted behaviour.

def find_weight(texel, w, windex):
    """Returns position *i* at which weight *windex* switches to value *w*."""
    assert type(w) is int
    if w == 0:
        return 0
    if provides_childs(texel):
        sum_w = 0
        for i1, i2, child in iter_childs(texel):
            delta = child.weights[windex]
            if sum_w+delta >= w:
                return find_weight(child, w-sum_w, windex)+i1
            sum_w += delta
    if w == texel.weights[windex]:
        return length(texel)
    raise NotFound(w)


def next_change(texel, windex, i):
    """Returns the next position from *i* at which weight *windex* changes."""
    if provides_childs(texel):
        for j1, j2, child in iter_childs(texel):
            if i1 < j2 and j1 < i2: # intersection
                n = next_change(texel, windex, i-j1)
                if n is not None:
                    return n+j1
    if texel.weights[windex]:
        return 0


def get_weight(texel, windex, i): 
    """Returns the weight *windex* at index *i*.
       pre:
         isinstance(texel, Texel)
    """
    if i<0:
        raise IndexError
    if i >= length(texel):
        return texel.weights[windex]
    w = 0
    if i == 0:
        return w

    if provides_childs(texel):
        for i1, i2, child in iter_childs(texel):
            if i2 <= i:
                w += child.weights[windex]
            elif i1 <= i <= i2:
                w += get_weight(child, windex, i-i1)
            else:
                break
    return w


if debug: # enable contract checking
     import contract
     contract.checkmod(__name__)



# -*- coding: latin-1 -*-


from .textmodel.textmodel import TextModel, dump_range
from .textmodel.texeltree import Texel, T, G, NL, Container, Single, \
    iter_childs, dump, NULL_TEXEL, length, grouped, copy

import wx
import StringIO
import base64    



def mk_textmodel(texel):
    model = TextModel()
    model.texel = texel
    return model



class Cell(Container):
    pass


class TextCell(Cell):
    text = property(lambda s:s.childs[1])
    def __init__(self, text, **kwargs):
        assert isinstance(text, Texel)
        self.childs = [NL, text, NL]
        self.compute_weights()



class ScriptingCell(Cell):

    client_name = 'direct python'

    input = property(lambda s:s.childs[1])
    output = property(lambda s:s.childs[3])

    def __init__(self, input, output, number = 0, **kwargs):
        assert isinstance(input, Texel)
        assert isinstance(output, Texel)
        self.number = number
        self.childs = [NL, input, NL, output, NL]
        self.compute_weights()


def strip_output(texel):
    if isinstance(texel, ScriptingCell):
        clone = texel.__class__(texel.input, NULL_TEXEL)
        return clone
    elif texel.is_group:
        r = []
        for child in texel.childs:
            r.append(strip_output(child))
        return G(r)
    else: 
        return texel


def _split_cell(texel, i):
    if i<=0 or i>=length(texel): return [texel]
    if texel.is_group:
        r = []
        for i1, i2, child in iter_childs(texel):
            if i1<i<i2:
                r.append(split_cell(child, i-i1))
            else:
                r.append(child)
        return r
    if isinstance(texel, Cell):
        r1 = []
        r2 = []
        for i1, i2, child in iter_childs(texel):
            if i1<i<i2:
                j = i-i1
                r1.append(grouped(copy(child, 0, j)))
                r2.append(grouped(copy(child, j, i2-i1)))
            else:
                r1.append(child)
                r2.append(child)
        return [strip_output(texel.set_childs(r1)),
                texel.set_childs(r2)]        
    return [texel]

def split_cell(texel, i):
    return grouped(_split_cell(texel, i))


class NotFound(Exception): pass

def find_cell(texel, i, i0=0):
    if isinstance(texel, Cell):
        return i0, texel
    elif texel.is_group:
        for j1, j2, child in iter_childs(texel):
            if j1<=i<j2:
                return find_cell(child, i-j1, i0+j1)
    raise NotFound()


def _bitmap_saver(bitmap):
    # we convert images to png before saving to save disk space
    w, h = bitmap.size
    im = wx.ImageFromData(w, h, bitmap.data)
    if isinstance(bitmap, BitmapRGBA):
        im.SetAlphaBuffer(bitmap.alpha)
    output = StringIO.StringIO()
    im.SaveStream(output, wx.BITMAP_TYPE_PNG)

    r = base64.b64encode(output.getvalue())
    return base64.b64encode(output.getvalue())



def _bitmap_loader(data):
    pngdata = base64.b64decode(data)
    stream = StringIO.StringIO(pngdata)
    image = wx.ImageFromStream(stream, type=wx.BITMAP_TYPE_ANY)
    return (image.Width, image.Height), image.Data, image.AlphaData



class BitmapRGB(Single):
    def __init__(self, data, size):
        self.data = data
        self.size = size

    def __repr__(self):
        return 'BitmapRGB(...)'

    def __getstate__(self):
        return _bitmap_saver(self)

    def __setstate__(self, data):
        self.size, self.data = _bitmap_loader(data)[:2]

    def draw(self, gc, state):
        w, h = self.size
        bitmap = wx.BitmapFromBuffer(w, h, self.data)
        m = gc.CreateMatrix()
        gc.SetTransform(m)
        gc.ConcatTransform(state['trafo'])
        gc.ConcatTransform(state['matrix'])
        gc.DrawBitmap(bitmap, 0, 0, w, h)
        gc.SetTransform(state['trafo'])


class BitmapRGBA(Single):
    def __init__(self, data, alpha, size):
        self.data = data
        self.alpha = alpha
        self.size = size

    def __repr__(self):
        return 'BitmapRGB(...)'

    def __getstate__(self):
        return _bitmap_saver(self)

    def __setstate__(self, data):
        self.size, self.data, self.alpha = _bitmap_loader(data)

    def draw(self, gc, state):
        w, h = self.size
        im = wx.ImageFromData(w, h, self.data)
        im.SetAlphaBuffer(self.alpha)
        bitmap = wx.BitmapFromImage(im)
        m = gc.CreateMatrix()
        gc.SetTransform(m)
        gc.ConcatTransform(state['trafo'])
        gc.ConcatTransform(state['matrix'])
        gc.DrawBitmap(bitmap, 0, 0, w, h)
        gc.SetTransform(state['trafo'])


class Graphics(Single):
    def __init__(self, item, size=(100, 100), frame=False):
        if type(item) is tuple:
            self.items = item
        elif type(item) is list:
            self.items = tuple(item)
        else:
            self.items = (item)
        self.size = size
        self.frame = frame



def test_00():
    "TextCell"
    cell = TextCell(T('0123456789'))
    model = TextModel()
    cellmodel = mk_textmodel(cell)
    model.insert(0, cellmodel)
    dump(model.texel)
    model.insert_text(4, "xyz")
    dump(model.texel)


def test_01():
    import pickle
    app = wx.App()

    im = wx.ArtProvider.GetBitmap(wx.ART_WARNING, size=(128, 128)).ConvertToImage()
    texel = BitmapRGBA(im.GetData(), im.GetAlphaData(), im.GetSize())
    s = pickle.dumps(texel)
    obj = pickle.loads(s)
    s2 = pickle.dumps(obj)
    assert s == s2

    texel = BitmapRGB(im.GetData(), im.GetSize())
    s = pickle.dumps(texel)
    obj = pickle.loads(s)
    s2 = pickle.dumps(obj)
    assert s == s2



# -*- coding: latin-1 -*-

from .testdevice import TESTDEVICE
from .boxes import TabulatorBox, TextBox, EmptyTextBox, NewlineBox, Row



def find_goodbreak(box, w):
    # Try to break after a space. Otherwise, return None.
    if not isinstance(box, TextBox):
        return 0
    text = box.text
    parts = box.measure_parts(text)

    for i, part in reversed(tuple(enumerate(parts))):
        if i>0 and part <= w and text[i] == ' ':
            return i+1
    return None


def find_anybreak(box, w):
    if not isinstance(box, TextBox):
        return 0
    text = box.text
    parts = box.measure_parts(text)

    for i, part in enumerate(parts):
        if part > w:
            return i
    return None


def split_box(box, i):
    if not isinstance(box, TextBox):
        assert i == 0
        return EmptyTextBox(), box
    text = box.text
    style = box.style
    device = box.device
    b1 = box.__class__(text[:i], style, device)
    b2 = box.__class__(text[i:], style, device)
    return b1, b2


def simple_linewrap(boxes, maxw, tabstops=(), wordwrap=True, 
                    device=TESTDEVICE):
    assert isinstance(maxw, int)
    l = []
    rows = [l]
    w = 0
    boxes = list(boxes[:])
    last = None
    while boxes:
        box = boxes[0]
        boxes = boxes[1:]
        if not len(box):
            continue
        if w+box.width <= maxw:
            l.append(box)
            w += box.width
            if isinstance(box, TextBox) and ' ' in box.text:
                i = box.text.rindex(' ')+1
                last = len(l)-1, i
            continue

        # start a new line
        split_at_i = True
        i = None
        if wordwrap:
            i = find_goodbreak(box, maxw-w)
        else:
            i = find_anybreak(box, maxw-w)
        if i is None:
            if last:
                k, j = last
                lastbox = l[k]
                a, b = split_box(lastbox, j)
                assert len(a)+len(b) == len(lastbox)
                boxes = [b]+l[k+1:]+[box]+boxes
                del l[k:]            
                l.append(a)
                split_at_i = False
            else:
                i = find_anybreak(box, maxw-w)

        if split_at_i:
            if i == 0:
                if len(l):
                    boxes = [box]+boxes
                else:
                    l.append(box)
            else:
                a, b = split_box(box, i)
                l.append(a)                    
                boxes = [b]+boxes
            
        # start a new line
        w = 0
        l = []
        rows.append(l)
        last = None
        
    # Remove the last row, if it is empty. 
    if not rows[-1]:
        rows = rows[:-1]
    return [Row(l, device) for l in rows]


def test_00():
    "find_break"
    box = TextBox("123 567 90")
    assert find_goodbreak(box, 3) == None
    assert find_goodbreak(box, 4) == 4
    assert find_goodbreak(box, 5) == 4
    assert find_goodbreak(box, 6) == 4
    assert find_goodbreak(box, 7) == 4
    assert find_goodbreak(box, 8) == 8
    assert find_goodbreak(box, 9) == 8
    assert find_goodbreak(box, 10) == 8
    assert find_goodbreak(box, 11) == 8 # XXX Hmm?


def test_01():
    boxes = []
    for text in "aa bb cc dd ee".split():
        boxes.append(TextBox(text))
        if text == 'dd':
            boxes.append(NewlineBox())

    assert str(simple_linewrap(boxes, 5)) == \
        "[Row[TB('aa'), TB('bb'), TB('c')], Row[TB('c'), TB('dd'), NL, "\
        "TB('ee')]]"

    boxes = []
    for text in "ff gg_hh ii jj".split('_'):
        boxes.append(TextBox(text))
    print str(simple_linewrap(boxes, 5))
    assert str(simple_linewrap(boxes, 5)) == \
        "[Row[TB('ff ')], Row[TB('gg'), TB('hh ')], Row[TB('ii jj')]]"



# -*- coding: latin-1 -*-



def pyshell(namespace=None):
    import wx.py
    style = wx.DEFAULT_FRAME_STYLE 
    f = wx.Frame(None,
                 pos=wx.DefaultPosition,
                 style=style
                 )
    f.Bind(wx.EVT_CLOSE, lambda e: f.Hide())
    box = wx.BoxSizer(wx.HORIZONTAL)
    if namespace is None:
        namespace = {}
    shell = wx.py.shell.Shell(
        f, -1, locals=namespace,                  
        size=(1000, 200),
        )
    f.shell = shell
    box.Add(shell, 1, wx.EXPAND)
    f.SetSizer(box)
    box.Fit(f)
    f.Fit()
    f.Show()




    



# -*- coding: latin-1 -*-


from setuptools import setup

# NOTE: if there is a problem with symlinks, update setuptools:
# http://stackoverflow.com/questions/27459888/how-to-make-setuptools-follow-symlinks

with open('README') as file:
    long_description = file.read()

setup(name='pynotebook',
      version='0.2.6',
      description = 'A wxPython based notebook environment for interactive computing.',
      long_description = long_description,
      author='C. Ecker',
      author_email='textmodelview@gmail.com',
      url='https://pypi.python.org/pypi/pynotebook/',
      license='BSD',
      packages=['pynotebook', 'pynotebook/textmodel', 'pynotebook/wxtextview'],
      platforms = "Linux, Mac OS X, Windows",
      keywords = ['Interactive', 'Interpreter', 'Notebook', 'Shell', 'wxPython'],
      classifiers = [
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
      ],
     )

